from __future__ import unicode_literals

from decimal import Decimal

import boto
import boto3
from boto3.dynamodb.conditions import Key
from botocore.exceptions import ClientError
import sure  # noqa
from freezegun import freeze_time
from moto import mock_dynamodb2, mock_dynamodb2_deprecated
from boto.exception import JSONResponseError
from tests.helpers import requires_boto_gte
try:
    from boto.dynamodb2.fields import GlobalAllIndex, HashKey, RangeKey, AllIndex
    from boto.dynamodb2.table import Item, Table
    from boto.dynamodb2.types import STRING, NUMBER
    from boto.dynamodb2.exceptions import ValidationException
    from boto.dynamodb2.exceptions import ConditionalCheckFailedException
except ImportError:
    pass


def create_table():
    table = Table.create('messages', schema=[
        HashKey('forum_name'),
        RangeKey('subject'),
    ], throughput={
        'read': 10,
        'write': 10,
    })
    return table


def create_table_with_local_indexes():
    table = Table.create(
        'messages',
        schema=[
            HashKey('forum_name'),
            RangeKey('subject'),
        ],
        throughput={
            'read': 10,
            'write': 10,
        },
        indexes=[
            AllIndex(
                'threads_index',
                parts=[
                    HashKey('forum_name', data_type=STRING),
                    RangeKey('threads', data_type=NUMBER),
                ]
            )
        ]
    )
    return table


def iterate_results(res):
    for i in res:
        pass


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
@freeze_time("2012-01-14")
def test_create_table():
    table = create_table()
    expected = {
        'Table': {
            'AttributeDefinitions': [
                {'AttributeName': 'forum_name', 'AttributeType': 'S'},
                {'AttributeName': 'subject', 'AttributeType': 'S'}
            ],
            'ProvisionedThroughput': {
                'NumberOfDecreasesToday': 0, 'WriteCapacityUnits': 10, 'ReadCapacityUnits': 10
            },
            'TableSizeBytes': 0,
            'TableName': 'messages',
            'TableStatus': 'ACTIVE',
            'TableArn': 'arn:aws:dynamodb:us-east-1:123456789011:table/messages',
            'KeySchema': [
                {'KeyType': 'HASH', 'AttributeName': 'forum_name'},
                {'KeyType': 'RANGE', 'AttributeName': 'subject'}
            ],
            'LocalSecondaryIndexes': [],
            'ItemCount': 0, 'CreationDateTime': 1326499200.0,
            'GlobalSecondaryIndexes': []
        }
    }
    table.describe().should.equal(expected)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
@freeze_time("2012-01-14")
def test_create_table_with_local_index():
    table = create_table_with_local_indexes()
    expected = {
        'Table': {
            'AttributeDefinitions': [
                {'AttributeName': 'forum_name', 'AttributeType': 'S'},
                {'AttributeName': 'subject', 'AttributeType': 'S'},
                {'AttributeName': 'threads', 'AttributeType': 'N'}
            ],
            'ProvisionedThroughput': {
                'NumberOfDecreasesToday': 0,
                'WriteCapacityUnits': 10,
                'ReadCapacityUnits': 10,
            },
            'TableSizeBytes': 0,
            'TableName': 'messages',
            'TableStatus': 'ACTIVE',
            'TableArn': 'arn:aws:dynamodb:us-east-1:123456789011:table/messages',
            'KeySchema': [
                {'KeyType': 'HASH', 'AttributeName': 'forum_name'},
                {'KeyType': 'RANGE', 'AttributeName': 'subject'}
            ],
            'LocalSecondaryIndexes': [
                {
                    'IndexName': 'threads_index',
                    'KeySchema': [
                        {'AttributeName': 'forum_name', 'KeyType': 'HASH'},
                        {'AttributeName': 'threads', 'KeyType': 'RANGE'}
                    ],
                    'Projection': {'ProjectionType': 'ALL'}
                }
            ],
            'ItemCount': 0,
            'CreationDateTime': 1326499200.0,
            'GlobalSecondaryIndexes': []
        }
    }
    table.describe().should.equal(expected)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_delete_table():
    conn = boto.dynamodb2.layer1.DynamoDBConnection()
    table = create_table()
    conn.list_tables()["TableNames"].should.have.length_of(1)

    table.delete()
    conn.list_tables()["TableNames"].should.have.length_of(0)
    conn.delete_table.when.called_with(
        'messages').should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_update_table_throughput():
    table = create_table()
    table.throughput["read"].should.equal(10)
    table.throughput["write"].should.equal(10)
    table.update(throughput={
        'read': 5,
        'write': 15,
    })

    table.throughput["read"].should.equal(5)
    table.throughput["write"].should.equal(15)

    table.update(throughput={
        'read': 5,
        'write': 6,
    })

    table.describe()

    table.throughput["read"].should.equal(5)
    table.throughput["write"].should.equal(6)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_item_add_and_describe_and_update():
    table = create_table()
    ok = table.put_item(data={
        'forum_name': 'LOLCat Forum',
        'subject': 'Check this out!',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    })
    ok.should.equal(True)

    table.get_item(forum_name="LOLCat Forum",
                   subject='Check this out!').should_not.be.none

    returned_item = table.get_item(
        forum_name='LOLCat Forum',
        subject='Check this out!'
    )
    dict(returned_item).should.equal({
        'forum_name': 'LOLCat Forum',
        'subject': 'Check this out!',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    })

    returned_item['SentBy'] = 'User B'
    returned_item.save(overwrite=True)

    returned_item = table.get_item(
        forum_name='LOLCat Forum',
        subject='Check this out!'
    )
    dict(returned_item).should.equal({
        'forum_name': 'LOLCat Forum',
        'subject': 'Check this out!',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User B',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    })


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_item_partial_save():
    table = create_table()

    data = {
        'forum_name': 'LOLCat Forum',
        'subject': 'The LOLz',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
    }

    table.put_item(data=data)
    returned_item = table.get_item(
        forum_name="LOLCat Forum", subject='The LOLz')

    returned_item['SentBy'] = 'User B'
    returned_item.partial_save()

    returned_item = table.get_item(
        forum_name='LOLCat Forum',
        subject='The LOLz'
    )
    dict(returned_item).should.equal({
        'forum_name': 'LOLCat Forum',
        'subject': 'The LOLz',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User B',
    })


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_item_put_without_table():
    table = Table('undeclared-table')
    item_data = {
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    }
    item = Item(table, item_data)
    item.save.when.called_with().should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_get_missing_item():
    table = create_table()

    table.get_item.when.called_with(
        hash_key='tester',
        range_key='other',
    ).should.throw(ValidationException)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_get_item_with_undeclared_table():
    table = Table('undeclared-table')
    table.get_item.when.called_with(
        test_hash=3241526475).should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_get_item_without_range_key():
    table = Table.create('messages', schema=[
        HashKey('test_hash'),
        RangeKey('test_range'),
    ], throughput={
        'read': 10,
        'write': 10,
    })

    hash_key = 3241526475
    range_key = 1234567890987
    table.put_item(data={'test_hash': hash_key, 'test_range': range_key})
    table.get_item.when.called_with(
        test_hash=hash_key).should.throw(ValidationException)


@requires_boto_gte("2.30.0")
@mock_dynamodb2_deprecated
def test_delete_item():
    table = create_table()
    item_data = {
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    }
    item = Item(table, item_data)
    item['subject'] = 'Check this out!'
    item.save()
    table.count().should.equal(1)

    response = item.delete()
    response.should.equal(True)

    table.count().should.equal(0)
    # Deletes are idempotent
    item.delete().should.equal(True)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_delete_item_with_undeclared_table():
    table = Table("undeclared-table")
    item_data = {
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    }
    item = Item(table, item_data)
    item.delete.when.called_with().should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query():
    table = create_table()

    item_data = {
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
        'subject': 'Check this out!'
    }
    item = Item(table, item_data)
    item.save(overwrite=True)

    item['forum_name'] = 'the-key'
    item['subject'] = '456'
    item.save(overwrite=True)

    item['forum_name'] = 'the-key'
    item['subject'] = '123'
    item.save(overwrite=True)

    item['forum_name'] = 'the-key'
    item['subject'] = '789'
    item.save(overwrite=True)

    table.count().should.equal(4)

    results = table.query_2(forum_name__eq='the-key',
                            subject__gt='1', consistent=True)
    expected = ["123", "456", "789"]
    for index, item in enumerate(results):
        item["subject"].should.equal(expected[index])

    results = table.query_2(forum_name__eq="the-key",
                            subject__gt='1', reverse=True)
    for index, item in enumerate(results):
        item["subject"].should.equal(expected[len(expected) - 1 - index])

    results = table.query_2(forum_name__eq='the-key',
                            subject__gt='1', consistent=True)
    sum(1 for _ in results).should.equal(3)

    results = table.query_2(forum_name__eq='the-key',
                            subject__gt='234', consistent=True)
    sum(1 for _ in results).should.equal(2)

    results = table.query_2(forum_name__eq='the-key', subject__gt='9999')
    sum(1 for _ in results).should.equal(0)

    results = table.query_2(forum_name__eq='the-key', subject__beginswith='12')
    sum(1 for _ in results).should.equal(1)

    results = table.query_2(forum_name__eq='the-key', subject__beginswith='7')
    sum(1 for _ in results).should.equal(1)

    results = table.query_2(forum_name__eq='the-key',
                            subject__between=['567', '890'])
    sum(1 for _ in results).should.equal(1)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query_with_undeclared_table():
    table = Table('undeclared')
    results = table.query(
        forum_name__eq='Amazon DynamoDB',
        subject__beginswith='DynamoDB',
        limit=1
    )
    iterate_results.when.called_with(results).should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_scan():
    table = create_table()
    item_data = {
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    }
    item_data['forum_name'] = 'the-key'
    item_data['subject'] = '456'

    item = Item(table, item_data)
    item.save()

    item['forum_name'] = 'the-key'
    item['subject'] = '123'
    item.save()

    item_data = {
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User B',
        'ReceivedTime': '12/9/2011 11:36:09 PM',
        'Ids': set([1, 2, 3]),
        'PK': 7,
    }

    item_data['forum_name'] = 'the-key'
    item_data['subject'] = '789'

    item = Item(table, item_data)
    item.save()

    results = table.scan()
    sum(1 for _ in results).should.equal(3)

    results = table.scan(SentBy__eq='User B')
    sum(1 for _ in results).should.equal(1)

    results = table.scan(Body__beginswith='http')
    sum(1 for _ in results).should.equal(3)

    results = table.scan(Ids__null=False)
    sum(1 for _ in results).should.equal(1)

    results = table.scan(Ids__null=True)
    sum(1 for _ in results).should.equal(2)

    results = table.scan(PK__between=[8, 9])
    sum(1 for _ in results).should.equal(0)

    results = table.scan(PK__between=[5, 8])
    sum(1 for _ in results).should.equal(1)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_scan_with_undeclared_table():
    conn = boto.dynamodb2.layer1.DynamoDBConnection()
    conn.scan.when.called_with(
        table_name='undeclared-table',
        scan_filter={
            "SentBy": {
                "AttributeValueList": [{
                    "S": "User B"}
                ],
                "ComparisonOperator": "EQ"
            }
        },
    ).should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_write_batch():
    table = create_table()
    with table.batch_write() as batch:
        batch.put_item(data={
            'forum_name': 'the-key',
            'subject': '123',
            'Body': 'http://url_to_lolcat.gif',
            'SentBy': 'User A',
            'ReceivedTime': '12/9/2011 11:36:03 PM',
        })
        batch.put_item(data={
            'forum_name': 'the-key',
            'subject': '789',
            'Body': 'http://url_to_lolcat.gif',
            'SentBy': 'User B',
            'ReceivedTime': '12/9/2011 11:36:03 PM',
        })

    table.count().should.equal(2)
    with table.batch_write() as batch:
        batch.delete_item(
            forum_name='the-key',
            subject='789'
        )

    table.count().should.equal(1)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_batch_read():
    table = create_table()
    item_data = {
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    }

    item_data['forum_name'] = 'the-key'
    item_data['subject'] = '456'

    item = Item(table, item_data)
    item.save()

    item = Item(table, item_data)
    item_data['forum_name'] = 'the-key'
    item_data['subject'] = '123'
    item.save()

    item_data = {
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User B',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
        'Ids': set([1, 2, 3]),
        'PK': 7,
    }
    item = Item(table, item_data)
    item_data['forum_name'] = 'another-key'
    item_data['subject'] = '789'
    item.save()
    results = table.batch_get(
        keys=[
            {'forum_name': 'the-key', 'subject': '123'},
            {'forum_name': 'another-key', 'subject': '789'},
        ]
    )

    # Iterate through so that batch_item gets called
    count = len([x for x in results])
    count.should.equal(2)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_get_key_fields():
    table = create_table()
    kf = table.get_key_fields()
    kf.should.equal(['forum_name', 'subject'])


@mock_dynamodb2_deprecated
def test_create_with_global_indexes():
    conn = boto.dynamodb2.layer1.DynamoDBConnection()

    Table.create('messages', schema=[
        HashKey('subject'),
        RangeKey('version'),
    ], global_indexes=[
        GlobalAllIndex('topic-created_at-index',
                       parts=[
                           HashKey('topic'),
                           RangeKey('created_at', data_type='N')
                       ],
                       throughput={
                           'read': 6,
                           'write': 1
                       }
                       ),
    ])

    table_description = conn.describe_table("messages")
    table_description['Table']["GlobalSecondaryIndexes"].should.equal([
        {
            "IndexName": "topic-created_at-index",
            "KeySchema": [
                {
                    "AttributeName": "topic",
                    "KeyType": "HASH"
                },
                {
                    "AttributeName": "created_at",
                    "KeyType": "RANGE"
                },
            ],
            "Projection": {
                "ProjectionType": "ALL"
            },
            "ProvisionedThroughput": {
                "ReadCapacityUnits": 6,
                "WriteCapacityUnits": 1,
            }
        }
    ])


@mock_dynamodb2_deprecated
def test_query_with_global_indexes():
    table = Table.create('messages', schema=[
        HashKey('subject'),
        RangeKey('version'),
    ], global_indexes=[
        GlobalAllIndex('topic-created_at-index',
                       parts=[
                           HashKey('topic'),
                           RangeKey('created_at', data_type='N')
                       ],
                       throughput={
                           'read': 6,
                           'write': 1
                       }
                       ),
        GlobalAllIndex('status-created_at-index',
                       parts=[
                           HashKey('status'),
                           RangeKey('created_at', data_type='N')
                       ],
                       throughput={
                           'read': 2,
                           'write': 1
                       }
                       )
    ])

    item_data = {
        'subject': 'Check this out!',
        'version': '1',
        'created_at': 0,
        'status': 'inactive'
    }
    item = Item(table, item_data)
    item.save(overwrite=True)

    item['version'] = '2'
    item.save(overwrite=True)

    results = table.query(status__eq='active')
    list(results).should.have.length_of(0)


@mock_dynamodb2_deprecated
def test_query_with_local_indexes():
    table = create_table_with_local_indexes()
    item_data = {
        'forum_name': 'Cool Forum',
        'subject': 'Check this out!',
        'version': '1',
        'threads': 1,
        'status': 'inactive'
    }
    item = Item(table, item_data)
    item.save(overwrite=True)

    item['version'] = '2'
    item.save(overwrite=True)
    results = table.query(forum_name__eq='Cool Forum',
                          index='threads_index', threads__eq=1)
    list(results).should.have.length_of(1)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query_filter_eq():
    table = create_table_with_local_indexes()
    item_data = [
        {
            'forum_name': 'Cool Forum',
            'subject': 'Check this out!',
            'version': '1',
            'threads': 1,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Read this now!',
            'version': '1',
            'threads': 5,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Please read this... please',
            'version': '1',
            'threads': 0,
        }
    ]
    for data in item_data:
        item = Item(table, data)
        item.save(overwrite=True)
    results = table.query_2(
        forum_name__eq='Cool Forum', index='threads_index', threads__eq=5
    )
    list(results).should.have.length_of(1)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query_filter_lt():
    table = create_table_with_local_indexes()
    item_data = [
        {
            'forum_name': 'Cool Forum',
            'subject': 'Check this out!',
            'version': '1',
            'threads': 1,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Read this now!',
            'version': '1',
            'threads': 5,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Please read this... please',
            'version': '1',
            'threads': 0,
        }
    ]
    for data in item_data:
        item = Item(table, data)
        item.save(overwrite=True)

    results = table.query(
        forum_name__eq='Cool Forum', index='threads_index', threads__lt=5
    )
    results = list(results)
    results.should.have.length_of(2)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query_filter_gt():
    table = create_table_with_local_indexes()
    item_data = [
        {
            'forum_name': 'Cool Forum',
            'subject': 'Check this out!',
            'version': '1',
            'threads': 1,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Read this now!',
            'version': '1',
            'threads': 5,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Please read this... please',
            'version': '1',
            'threads': 0,
        }
    ]
    for data in item_data:
        item = Item(table, data)
        item.save(overwrite=True)

    results = table.query(
        forum_name__eq='Cool Forum', index='threads_index', threads__gt=1
    )
    list(results).should.have.length_of(1)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query_filter_lte():
    table = create_table_with_local_indexes()
    item_data = [
        {
            'forum_name': 'Cool Forum',
            'subject': 'Check this out!',
            'version': '1',
            'threads': 1,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Read this now!',
            'version': '1',
            'threads': 5,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Please read this... please',
            'version': '1',
            'threads': 0,
        }
    ]
    for data in item_data:
        item = Item(table, data)
        item.save(overwrite=True)

    results = table.query(
        forum_name__eq='Cool Forum', index='threads_index', threads__lte=5
    )
    list(results).should.have.length_of(3)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query_filter_gte():
    table = create_table_with_local_indexes()
    item_data = [
        {
            'forum_name': 'Cool Forum',
            'subject': 'Check this out!',
            'version': '1',
            'threads': 1,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Read this now!',
            'version': '1',
            'threads': 5,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Please read this... please',
            'version': '1',
            'threads': 0,
        }
    ]
    for data in item_data:
        item = Item(table, data)
        item.save(overwrite=True)

    results = table.query(
        forum_name__eq='Cool Forum', index='threads_index', threads__gte=1
    )
    list(results).should.have.length_of(2)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query_non_hash_range_key():
    table = create_table_with_local_indexes()
    item_data = [
        {
            'forum_name': 'Cool Forum',
            'subject': 'Check this out!',
            'version': '1',
            'threads': 1,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Read this now!',
            'version': '3',
            'threads': 5,
        },
        {
            'forum_name': 'Cool Forum',
            'subject': 'Please read this... please',
            'version': '2',
            'threads': 0,
        }
    ]
    for data in item_data:
        item = Item(table, data)
        item.save(overwrite=True)

    results = table.query(
        forum_name__eq='Cool Forum', version__gt="2"
    )
    results = list(results)
    results.should.have.length_of(1)

    results = table.query(
        forum_name__eq='Cool Forum', version__lt="3"
    )
    results = list(results)
    results.should.have.length_of(2)


@mock_dynamodb2_deprecated
def test_reverse_query():
    conn = boto.dynamodb2.layer1.DynamoDBConnection()

    table = Table.create('messages', schema=[
        HashKey('subject'),
        RangeKey('created_at', data_type='N')
    ])

    for i in range(10):
        table.put_item({
            'subject': "Hi",
            'created_at': i
        })

    results = table.query_2(subject__eq="Hi",
                            created_at__lt=6,
                            limit=4,
                            reverse=True)

    expected = [Decimal(5), Decimal(4), Decimal(3), Decimal(2)]
    [r['created_at'] for r in results].should.equal(expected)


@mock_dynamodb2_deprecated
def test_lookup():
    from decimal import Decimal
    table = Table.create('messages', schema=[
        HashKey('test_hash'),
        RangeKey('test_range'),
    ], throughput={
        'read': 10,
        'write': 10,
    })

    hash_key = 3241526475
    range_key = 1234567890987
    data = {'test_hash': hash_key, 'test_range': range_key}
    table.put_item(data=data)
    message = table.lookup(hash_key, range_key)
    message.get('test_hash').should.equal(Decimal(hash_key))
    message.get('test_range').should.equal(Decimal(range_key))


@mock_dynamodb2_deprecated
def test_failed_overwrite():
    table = Table.create('messages', schema=[
        HashKey('id'),
        RangeKey('range'),
    ], throughput={
        'read': 7,
        'write': 3,
    })

    data1 = {'id': '123', 'range': 'abc', 'data': '678'}
    table.put_item(data=data1)

    data2 = {'id': '123', 'range': 'abc', 'data': '345'}
    table.put_item(data=data2, overwrite=True)

    data3 = {'id': '123', 'range': 'abc', 'data': '812'}
    table.put_item.when.called_with(data=data3).should.throw(
        ConditionalCheckFailedException)

    returned_item = table.lookup('123', 'abc')
    dict(returned_item).should.equal(data2)

    data4 = {'id': '123', 'range': 'ghi', 'data': 812}
    table.put_item(data=data4)

    returned_item = table.lookup('123', 'ghi')
    dict(returned_item).should.equal(data4)


@mock_dynamodb2_deprecated
def test_conflicting_writes():
    table = Table.create('messages', schema=[
        HashKey('id'),
        RangeKey('range'),
    ])

    item_data = {'id': '123', 'range': 'abc', 'data': '678'}
    item1 = Item(table, item_data)
    item2 = Item(table, item_data)
    item1.save()

    item1['data'] = '579'
    item2['data'] = '912'

    item1.save()
    item2.save.when.called_with().should.throw(ConditionalCheckFailedException)

"""
boto3
"""


@mock_dynamodb2
def test_boto3_conditions():
    dynamodb = boto3.resource('dynamodb', region_name='us-east-1')

    # Create the DynamoDB table.
    table = dynamodb.create_table(
        TableName='users',
        KeySchema=[
            {
                'AttributeName': 'forum_name',
                'KeyType': 'HASH'
            },
            {
                'AttributeName': 'subject',
                'KeyType': 'RANGE'
            },
        ],
        AttributeDefinitions=[
            {
                'AttributeName': 'forum_name',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'subject',
                'AttributeType': 'S'
            },
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 5
        }
    )
    table = dynamodb.Table('users')

    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '123'
    })
    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '456'
    })
    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '789'
    })

    # Test a query returning all items
    results = table.query(
        KeyConditionExpression=Key('forum_name').eq(
            'the-key') & Key("subject").gt('1'),
        ScanIndexForward=True,
    )
    expected = ["123", "456", "789"]
    for index, item in enumerate(results['Items']):
        item["subject"].should.equal(expected[index])

    # Return all items again, but in reverse
    results = table.query(
        KeyConditionExpression=Key('forum_name').eq(
            'the-key') & Key("subject").gt('1'),
        ScanIndexForward=False,
    )
    for index, item in enumerate(reversed(results['Items'])):
        item["subject"].should.equal(expected[index])

    # Filter the subjects to only return some of the results
    results = table.query(
        KeyConditionExpression=Key('forum_name').eq(
            'the-key') & Key("subject").gt('234'),
        ConsistentRead=True,
    )
    results['Count'].should.equal(2)

    # Filter to return no results
    results = table.query(
        KeyConditionExpression=Key('forum_name').eq(
            'the-key') & Key("subject").gt('9999')
    )
    results['Count'].should.equal(0)

    results = table.query(
        KeyConditionExpression=Key('forum_name').eq(
            'the-key') & Key("subject").begins_with('12')
    )
    results['Count'].should.equal(1)

    results = table.query(
        KeyConditionExpression=Key("subject").begins_with(
            '7') & Key('forum_name').eq('the-key')
    )
    results['Count'].should.equal(1)

    results = table.query(
        KeyConditionExpression=Key('forum_name').eq(
            'the-key') & Key("subject").between('567', '890')
    )
    results['Count'].should.equal(1)


@mock_dynamodb2
def test_boto3_put_item_with_conditions():
    import botocore
    dynamodb = boto3.resource('dynamodb', region_name='us-east-1')

    # Create the DynamoDB table.
    table = dynamodb.create_table(
        TableName='users',
        KeySchema=[
            {
                'AttributeName': 'forum_name',
                'KeyType': 'HASH'
            },
            {
                'AttributeName': 'subject',
                'KeyType': 'RANGE'
            },
        ],
        AttributeDefinitions=[
            {
                'AttributeName': 'forum_name',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'subject',
                'AttributeType': 'S'
            },
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 5
        }
    )
    table = dynamodb.Table('users')

    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '123'
    })

    table.put_item(
        Item={
            'forum_name': 'the-key-2',
            'subject': '1234',
        },
        ConditionExpression='attribute_not_exists(forum_name) AND attribute_not_exists(subject)'
    )

    table.put_item.when.called_with(
        Item={
            'forum_name': 'the-key',
            'subject': '123'
        },
        ConditionExpression='attribute_not_exists(forum_name) AND attribute_not_exists(subject)'
    ).should.throw(botocore.exceptions.ClientError)

    table.put_item.when.called_with(
        Item={
            'forum_name': 'bogus-key',
            'subject': 'bogus',
            'test': '123'
        },
        ConditionExpression='attribute_exists(forum_name) AND attribute_exists(subject)'
    ).should.throw(botocore.exceptions.ClientError)


def _create_table_with_range_key():
    dynamodb = boto3.resource('dynamodb', region_name='us-east-1')

    # Create the DynamoDB table.
    table = dynamodb.create_table(
        TableName='users',
        KeySchema=[
            {
                'AttributeName': 'forum_name',
                'KeyType': 'HASH'
            },
            {
                'AttributeName': 'subject',
                'KeyType': 'RANGE'
            },
        ],
        GlobalSecondaryIndexes=[{
            'IndexName': 'TestGSI',
            'KeySchema': [
                {
                    'AttributeName': 'username',
                    'KeyType': 'HASH',
                },
                {
                    'AttributeName': 'created',
                    'KeyType': 'RANGE',
                }
            ],
            'Projection': {
                'ProjectionType': 'ALL',
            },
            'ProvisionedThroughput': {
                'ReadCapacityUnits': 5,
                'WriteCapacityUnits': 5
            }
        }],
        AttributeDefinitions=[
            {
                'AttributeName': 'forum_name',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'subject',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'username',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'created',
                'AttributeType': 'N'
            }
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 5
        }
    )
    return dynamodb.Table('users')


@mock_dynamodb2
def test_update_item_range_key_set():
    table = _create_table_with_range_key()
    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '123',
        'username': 'johndoe',
        'created': Decimal('3'),
    })

    item_key = {'forum_name': 'the-key', 'subject': '123'}
    table.update_item(
        Key=item_key,
        AttributeUpdates={
            'username': {
                'Action': u'PUT',
                'Value': 'johndoe2'
            },
            'created': {
                'Action': u'PUT',
                'Value': Decimal('4'),
            },
            'mapfield': {
                'Action': u'PUT',
                'Value': {'key': 'value'},
            }
        },
    )

    returned_item = dict((k, str(v) if isinstance(v, Decimal) else v)
                         for k, v in table.get_item(Key=item_key)['Item'].items())
    dict(returned_item).should.equal({
        'username': "johndoe2",
        'forum_name': 'the-key',
        'subject': '123',
        'created': '4',
        'mapfield': {'key': 'value'},
    })


@mock_dynamodb2
def test_update_item_does_not_exist_is_created():
    table = _create_table_with_range_key()

    item_key = {'forum_name': 'the-key', 'subject': '123'}
    result = table.update_item(
        Key=item_key,
        AttributeUpdates={
            'username': {
                'Action': u'PUT',
                'Value': 'johndoe2'
            },
            'created': {
                'Action': u'PUT',
                'Value': Decimal('4'),
            },
            'mapfield': {
                'Action': u'PUT',
                'Value': {'key': 'value'},
            }
        },
        ReturnValues='ALL_OLD',
    )

    assert not result.get('Attributes')

    returned_item = dict((k, str(v) if isinstance(v, Decimal) else v)
                         for k, v in table.get_item(Key=item_key)['Item'].items())
    dict(returned_item).should.equal({
        'username': "johndoe2",
        'forum_name': 'the-key',
        'subject': '123',
        'created': '4',
        'mapfield': {'key': 'value'},
    })


@mock_dynamodb2
def test_update_item_add_value():
    table = _create_table_with_range_key()

    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '123',
        'numeric_field': Decimal('-1'),
    })

    item_key = {'forum_name': 'the-key', 'subject': '123'}
    table.update_item(
        Key=item_key,
        AttributeUpdates={
            'numeric_field': {
                'Action': u'ADD',
                'Value': Decimal('2'),
            },
        },
    )

    returned_item = dict((k, str(v) if isinstance(v, Decimal) else v)
                         for k, v in table.get_item(Key=item_key)['Item'].items())
    dict(returned_item).should.equal({
        'numeric_field': '1',
        'forum_name': 'the-key',
        'subject': '123',
    })


@mock_dynamodb2
def test_update_item_add_value_string_set():
    table = _create_table_with_range_key()

    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '123',
        'string_set': set(['str1', 'str2']),
    })

    item_key = {'forum_name': 'the-key', 'subject': '123'}
    table.update_item(
        Key=item_key,
        AttributeUpdates={
            'string_set': {
                'Action': u'ADD',
                'Value': set(['str3']),
            },
        },
    )

    returned_item = dict((k, str(v) if isinstance(v, Decimal) else v)
                         for k, v in table.get_item(Key=item_key)['Item'].items())
    dict(returned_item).should.equal({
        'string_set': set(['str1', 'str2', 'str3']),
        'forum_name': 'the-key',
        'subject': '123',
    })


@mock_dynamodb2
def test_update_item_add_value_does_not_exist_is_created():
    table = _create_table_with_range_key()

    item_key = {'forum_name': 'the-key', 'subject': '123'}
    table.update_item(
        Key=item_key,
        AttributeUpdates={
            'numeric_field': {
                'Action': u'ADD',
                'Value': Decimal('2'),
            },
        },
    )

    returned_item = dict((k, str(v) if isinstance(v, Decimal) else v)
                         for k, v in table.get_item(Key=item_key)['Item'].items())
    dict(returned_item).should.equal({
        'numeric_field': '2',
        'forum_name': 'the-key',
        'subject': '123',
    })


@mock_dynamodb2
def test_update_item_with_expression():
    table = _create_table_with_range_key()

    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '123',
        'field': '1'
    })

    item_key = {'forum_name': 'the-key', 'subject': '123'}

    table.update_item(
        Key=item_key,
        UpdateExpression='SET field=2',
    )
    dict(table.get_item(Key=item_key)['Item']).should.equal({
        'field': '2',
        'forum_name': 'the-key',
        'subject': '123',
    })

    table.update_item(
        Key=item_key,
        UpdateExpression='SET field  = 3',
    )
    dict(table.get_item(Key=item_key)['Item']).should.equal({
        'field': '3',
        'forum_name': 'the-key',
        'subject': '123',
    })

@mock_dynamodb2
def test_update_item_add_with_expression():
    table = _create_table_with_range_key()

    item_key = {'forum_name': 'the-key', 'subject': '123'}
    current_item = {
        'forum_name': 'the-key',
        'subject': '123',
        'str_set': {'item1', 'item2', 'item3'},
        'num_set': {1, 2, 3},
        'num_val': 6
    }

    # Put an entry in the DB to play with
    table.put_item(Item=current_item)

    # Update item to add a string value to a string set
    table.update_item(
        Key=item_key,
        UpdateExpression='ADD str_set :v',
        ExpressionAttributeValues={
            ':v': {'item4'}
        }
    )
    current_item['str_set'] = current_item['str_set'].union({'item4'})
    dict(table.get_item(Key=item_key)['Item']).should.equal(current_item)

    # Update item to add a num value to a num set
    table.update_item(
        Key=item_key,
        UpdateExpression='ADD num_set :v',
        ExpressionAttributeValues={
            ':v': {6}
        }
    )
    current_item['num_set'] = current_item['num_set'].union({6})
    dict(table.get_item(Key=item_key)['Item']).should.equal(current_item)

    # Update item to add a value to a number value
    table.update_item(
        Key=item_key,
        UpdateExpression='ADD num_val :v',
        ExpressionAttributeValues={
            ':v': 20
        }
    )
    current_item['num_val'] = current_item['num_val'] + 20
    dict(table.get_item(Key=item_key)['Item']).should.equal(current_item)

    # Attempt to add a number value to a string set, should raise Client Error
    table.update_item.when.called_with(
        Key=item_key,
        UpdateExpression='ADD str_set :v',
        ExpressionAttributeValues={
            ':v': 20
        }
    ).should.have.raised(ClientError)
    dict(table.get_item(Key=item_key)['Item']).should.equal(current_item)

    # Attempt to add a number set to the string set, should raise a ClientError
    table.update_item.when.called_with(
        Key=item_key,
        UpdateExpression='ADD str_set :v',
        ExpressionAttributeValues={
            ':v': { 20 }
        }
    ).should.have.raised(ClientError)
    dict(table.get_item(Key=item_key)['Item']).should.equal(current_item)

    # Attempt to update with a bad expression
    table.update_item.when.called_with(
        Key=item_key,
        UpdateExpression='ADD str_set bad_value'
    ).should.have.raised(ClientError)

    # Attempt to add a string value instead of a string set
    table.update_item.when.called_with(
        Key=item_key,
        UpdateExpression='ADD str_set :v',
        ExpressionAttributeValues={
            ':v': 'new_string'
        }
    ).should.have.raised(ClientError)


@mock_dynamodb2
def test_update_item_delete_with_expression():
    table = _create_table_with_range_key()

    item_key = {'forum_name': 'the-key', 'subject': '123'}
    current_item = {
        'forum_name': 'the-key',
        'subject': '123',
        'str_set': {'item1', 'item2', 'item3'},
        'num_set': {1, 2, 3},
        'num_val': 6
    }

    # Put an entry in the DB to play with
    table.put_item(Item=current_item)

    # Update item to delete a string value from a string set
    table.update_item(
        Key=item_key,
        UpdateExpression='DELETE str_set :v',
        ExpressionAttributeValues={
            ':v': {'item2'}
        }
    )
    current_item['str_set'] = current_item['str_set'].difference({'item2'})
    dict(table.get_item(Key=item_key)['Item']).should.equal(current_item)

    # Update item to delete  a num value from a num set
    table.update_item(
        Key=item_key,
        UpdateExpression='DELETE num_set :v',
        ExpressionAttributeValues={
            ':v': {2}
        }
    )
    current_item['num_set'] = current_item['num_set'].difference({2})
    dict(table.get_item(Key=item_key)['Item']).should.equal(current_item)

    # Try to delete on a number, this should fail
    table.update_item.when.called_with(
        Key=item_key,
        UpdateExpression='DELETE num_val :v',
        ExpressionAttributeValues={
            ':v': 20
        }
    ).should.have.raised(ClientError)
    dict(table.get_item(Key=item_key)['Item']).should.equal(current_item)

    # Try to delete a string set from a number set
    table.update_item.when.called_with(
        Key=item_key,
        UpdateExpression='DELETE num_set :v',
        ExpressionAttributeValues={
            ':v': {'del_str'}
        }
    ).should.have.raised(ClientError)
    dict(table.get_item(Key=item_key)['Item']).should.equal(current_item)

    # Attempt to update with a bad expression
    table.update_item.when.called_with(
        Key=item_key,
        UpdateExpression='DELETE num_val badvalue'
    ).should.have.raised(ClientError)


@mock_dynamodb2
def test_boto3_query_gsi_range_comparison():
    table = _create_table_with_range_key()

    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '123',
        'username': 'johndoe',
        'created': 3,
    })
    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '456',
        'username': 'johndoe',
        'created': 1,
    })
    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '789',
        'username': 'johndoe',
        'created': 2,
    })
    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '159',
        'username': 'janedoe',
        'created': 2,
    })
    table.put_item(Item={
        'forum_name': 'the-key',
        'subject': '601',
        'username': 'janedoe',
        'created': 5,
    })

    # Test a query returning all johndoe items
    results = table.query(
        KeyConditionExpression=Key('username').eq(
            'johndoe') & Key("created").gt(0),
        ScanIndexForward=True,
        IndexName='TestGSI',
    )
    expected = ["456", "789", "123"]
    for index, item in enumerate(results['Items']):
        item["subject"].should.equal(expected[index])

    # Return all johndoe items again, but in reverse
    results = table.query(
        KeyConditionExpression=Key('username').eq(
            'johndoe') & Key("created").gt(0),
        ScanIndexForward=False,
        IndexName='TestGSI',
    )
    for index, item in enumerate(reversed(results['Items'])):
        item["subject"].should.equal(expected[index])

    # Filter the creation to only return some of the results
    # And reverse order of hash + range key
    results = table.query(
        KeyConditionExpression=Key("created").gt(
            1) & Key('username').eq('johndoe'),
        ConsistentRead=True,
        IndexName='TestGSI',
    )
    results['Count'].should.equal(2)

    # Filter to return no results
    results = table.query(
        KeyConditionExpression=Key('username').eq(
            'janedoe') & Key("created").gt(9),
        IndexName='TestGSI',
    )
    results['Count'].should.equal(0)

    results = table.query(
        KeyConditionExpression=Key('username').eq(
            'janedoe') & Key("created").eq(5),
        IndexName='TestGSI',
    )
    results['Count'].should.equal(1)

    # Test range key sorting
    results = table.query(
        KeyConditionExpression=Key('username').eq(
            'johndoe') & Key("created").gt(0),
        IndexName='TestGSI',
    )
    expected = [Decimal('1'), Decimal('2'), Decimal('3')]
    for index, item in enumerate(results['Items']):
        item["created"].should.equal(expected[index])


@mock_dynamodb2
def test_boto3_update_table_throughput():
    dynamodb = boto3.resource('dynamodb', region_name='us-east-1')

    # Create the DynamoDB table.
    table = dynamodb.create_table(
        TableName='users',
        KeySchema=[
            {
                'AttributeName': 'forum_name',
                'KeyType': 'HASH'
            },
            {
                'AttributeName': 'subject',
                'KeyType': 'RANGE'
            },
        ],
        AttributeDefinitions=[
            {
                'AttributeName': 'forum_name',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'subject',
                'AttributeType': 'S'
            },
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 6
        }
    )
    table = dynamodb.Table('users')

    table.provisioned_throughput['ReadCapacityUnits'].should.equal(5)
    table.provisioned_throughput['WriteCapacityUnits'].should.equal(6)

    table.update(ProvisionedThroughput={
        'ReadCapacityUnits': 10,
        'WriteCapacityUnits': 11,
    })

    table = dynamodb.Table('users')

    table.provisioned_throughput['ReadCapacityUnits'].should.equal(10)
    table.provisioned_throughput['WriteCapacityUnits'].should.equal(11)


@mock_dynamodb2
def test_boto3_update_table_gsi_throughput():
    dynamodb = boto3.resource('dynamodb', region_name='us-east-1')

    # Create the DynamoDB table.
    table = dynamodb.create_table(
        TableName='users',
        KeySchema=[
            {
                'AttributeName': 'forum_name',
                'KeyType': 'HASH'
            },
            {
                'AttributeName': 'subject',
                'KeyType': 'RANGE'
            },
        ],
        GlobalSecondaryIndexes=[{
            'IndexName': 'TestGSI',
            'KeySchema': [
                {
                    'AttributeName': 'username',
                    'KeyType': 'HASH',
                },
                {
                    'AttributeName': 'created',
                    'KeyType': 'RANGE',
                }
            ],
            'Projection': {
                'ProjectionType': 'ALL',
            },
            'ProvisionedThroughput': {
                'ReadCapacityUnits': 3,
                'WriteCapacityUnits': 4
            }
        }],
        AttributeDefinitions=[
            {
                'AttributeName': 'forum_name',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'subject',
                'AttributeType': 'S'
            },
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 6
        }
    )
    table = dynamodb.Table('users')

    gsi_throughput = table.global_secondary_indexes[0]['ProvisionedThroughput']
    gsi_throughput['ReadCapacityUnits'].should.equal(3)
    gsi_throughput['WriteCapacityUnits'].should.equal(4)

    table.provisioned_throughput['ReadCapacityUnits'].should.equal(5)
    table.provisioned_throughput['WriteCapacityUnits'].should.equal(6)

    table.update(GlobalSecondaryIndexUpdates=[{
        'Update': {
            'IndexName': 'TestGSI',
            'ProvisionedThroughput': {
                'ReadCapacityUnits': 10,
                'WriteCapacityUnits': 11,
            }
        },
    }])

    table = dynamodb.Table('users')

    # Primary throughput has not changed
    table.provisioned_throughput['ReadCapacityUnits'].should.equal(5)
    table.provisioned_throughput['WriteCapacityUnits'].should.equal(6)

    gsi_throughput = table.global_secondary_indexes[0]['ProvisionedThroughput']
    gsi_throughput['ReadCapacityUnits'].should.equal(10)
    gsi_throughput['WriteCapacityUnits'].should.equal(11)


@mock_dynamodb2
def test_update_table_gsi_create():
    dynamodb = boto3.resource('dynamodb', region_name='us-east-1')

    # Create the DynamoDB table.
    table = dynamodb.create_table(
        TableName='users',
        KeySchema=[
            {
                'AttributeName': 'forum_name',
                'KeyType': 'HASH'
            },
            {
                'AttributeName': 'subject',
                'KeyType': 'RANGE'
            },
        ],
        AttributeDefinitions=[
            {
                'AttributeName': 'forum_name',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'subject',
                'AttributeType': 'S'
            },
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 6
        }
    )
    table = dynamodb.Table('users')

    table.global_secondary_indexes.should.have.length_of(0)

    table.update(GlobalSecondaryIndexUpdates=[{
        'Create': {
            'IndexName': 'TestGSI',
            'KeySchema': [
                {
                    'AttributeName': 'username',
                    'KeyType': 'HASH',
                },
                {
                    'AttributeName': 'created',
                    'KeyType': 'RANGE',
                }
            ],
            'Projection': {
                'ProjectionType': 'ALL',
            },
            'ProvisionedThroughput': {
                'ReadCapacityUnits': 3,
                'WriteCapacityUnits': 4
            }
        },
    }])

    table = dynamodb.Table('users')
    table.global_secondary_indexes.should.have.length_of(1)

    gsi_throughput = table.global_secondary_indexes[0]['ProvisionedThroughput']
    assert gsi_throughput['ReadCapacityUnits'].should.equal(3)
    assert gsi_throughput['WriteCapacityUnits'].should.equal(4)

    # Check update works
    table.update(GlobalSecondaryIndexUpdates=[{
        'Update': {
            'IndexName': 'TestGSI',
            'ProvisionedThroughput': {
                'ReadCapacityUnits': 10,
                'WriteCapacityUnits': 11,
            }
        },
    }])
    table = dynamodb.Table('users')

    gsi_throughput = table.global_secondary_indexes[0]['ProvisionedThroughput']
    assert gsi_throughput['ReadCapacityUnits'].should.equal(10)
    assert gsi_throughput['WriteCapacityUnits'].should.equal(11)

    table.update(GlobalSecondaryIndexUpdates=[{
        'Delete': {
            'IndexName': 'TestGSI',
        },
    }])

    table = dynamodb.Table('users')
    table.global_secondary_indexes.should.have.length_of(0)


@mock_dynamodb2
def test_update_table_gsi_throughput():
    dynamodb = boto3.resource('dynamodb', region_name='us-east-1')

    # Create the DynamoDB table.
    table = dynamodb.create_table(
        TableName='users',
        KeySchema=[
            {
                'AttributeName': 'forum_name',
                'KeyType': 'HASH'
            },
            {
                'AttributeName': 'subject',
                'KeyType': 'RANGE'
            },
        ],
        GlobalSecondaryIndexes=[{
            'IndexName': 'TestGSI',
            'KeySchema': [
                {
                    'AttributeName': 'username',
                    'KeyType': 'HASH',
                },
                {
                    'AttributeName': 'created',
                    'KeyType': 'RANGE',
                }
            ],
            'Projection': {
                'ProjectionType': 'ALL',
            },
            'ProvisionedThroughput': {
                'ReadCapacityUnits': 3,
                'WriteCapacityUnits': 4
            }
        }],
        AttributeDefinitions=[
            {
                'AttributeName': 'forum_name',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'subject',
                'AttributeType': 'S'
            },
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 6
        }
    )
    table = dynamodb.Table('users')
    table.global_secondary_indexes.should.have.length_of(1)

    table.update(GlobalSecondaryIndexUpdates=[{
        'Delete': {
            'IndexName': 'TestGSI',
        },
    }])

    table = dynamodb.Table('users')
    table.global_secondary_indexes.should.have.length_of(0)


@mock_dynamodb2
def test_query_pagination():
    table = _create_table_with_range_key()
    for i in range(10):
        table.put_item(Item={
            'forum_name': 'the-key',
            'subject': '{0}'.format(i),
            'username': 'johndoe',
            'created': Decimal('3'),
        })

    page1 = table.query(
        KeyConditionExpression=Key('forum_name').eq('the-key'),
        Limit=6
    )
    page1['Count'].should.equal(6)
    page1['Items'].should.have.length_of(6)
    page1.should.have.key('LastEvaluatedKey')

    page2 = table.query(
        KeyConditionExpression=Key('forum_name').eq('the-key'),
        Limit=6,
        ExclusiveStartKey=page1['LastEvaluatedKey']
    )
    page2['Count'].should.equal(4)
    page2['Items'].should.have.length_of(4)
    page2.should_not.have.key('LastEvaluatedKey')

    results = page1['Items'] + page2['Items']
    subjects = set([int(r['subject']) for r in results])
    subjects.should.equal(set(range(10)))



from __future__ import unicode_literals

import boto
from boto.exception import S3ResponseError
from boto.s3.lifecycle import Lifecycle, Transition, Expiration, Rule

import sure  # noqa

from moto import mock_s3_deprecated


@mock_s3_deprecated
def test_lifecycle_create():
    conn = boto.s3.connect_to_region("us-west-1")
    bucket = conn.create_bucket("foobar")

    lifecycle = Lifecycle()
    lifecycle.add_rule('myid', '', 'Enabled', 30)
    bucket.configure_lifecycle(lifecycle)
    response = bucket.get_lifecycle_config()
    len(response).should.equal(1)
    lifecycle = response[0]
    lifecycle.id.should.equal('myid')
    lifecycle.prefix.should.equal('')
    lifecycle.status.should.equal('Enabled')
    list(lifecycle.transition).should.equal([])


@mock_s3_deprecated
def test_lifecycle_with_glacier_transition():
    conn = boto.s3.connect_to_region("us-west-1")
    bucket = conn.create_bucket("foobar")

    lifecycle = Lifecycle()
    transition = Transition(days=30, storage_class='GLACIER')
    rule = Rule('myid', prefix='', status='Enabled', expiration=None,
                transition=transition)
    lifecycle.append(rule)
    bucket.configure_lifecycle(lifecycle)
    response = bucket.get_lifecycle_config()
    transition = response[0].transition
    transition.days.should.equal(30)
    transition.storage_class.should.equal('GLACIER')
    transition.date.should.equal(None)


@mock_s3_deprecated
def test_lifecycle_multi():
    conn = boto.s3.connect_to_region("us-west-1")
    bucket = conn.create_bucket("foobar")

    date = '2022-10-12T00:00:00.000Z'
    sc = 'GLACIER'
    lifecycle = Lifecycle()
    lifecycle.add_rule("1", "1/", "Enabled", 1)
    lifecycle.add_rule("2", "2/", "Enabled", Expiration(days=2))
    lifecycle.add_rule("3", "3/", "Enabled", Expiration(date=date))
    lifecycle.add_rule("4", "4/", "Enabled", None,
                       Transition(days=4, storage_class=sc))
    lifecycle.add_rule("5", "5/", "Enabled", None,
                       Transition(date=date, storage_class=sc))

    bucket.configure_lifecycle(lifecycle)
    # read the lifecycle back
    rules = bucket.get_lifecycle_config()

    for rule in rules:
        if rule.id == "1":
            rule.prefix.should.equal("1/")
            rule.expiration.days.should.equal(1)
        elif rule.id == "2":
            rule.prefix.should.equal("2/")
            rule.expiration.days.should.equal(2)
        elif rule.id == "3":
            rule.prefix.should.equal("3/")
            rule.expiration.date.should.equal(date)
        elif rule.id == "4":
            rule.prefix.should.equal("4/")
            rule.transition.days.should.equal(4)
            rule.transition.storage_class.should.equal(sc)
        elif rule.id == "5":
            rule.prefix.should.equal("5/")
            rule.transition.date.should.equal(date)
            rule.transition.storage_class.should.equal(sc)
        else:
            assert False, "Invalid rule id"


@mock_s3_deprecated
def test_lifecycle_delete():
    conn = boto.s3.connect_to_region("us-west-1")
    bucket = conn.create_bucket("foobar")

    lifecycle = Lifecycle()
    lifecycle.add_rule(expiration=30)
    bucket.configure_lifecycle(lifecycle)
    response = bucket.get_lifecycle_config()
    response.should.have.length_of(1)

    bucket.delete_lifecycle_configuration()
    bucket.get_lifecycle_config.when.called_with().should.throw(S3ResponseError)



from __future__ import unicode_literals

import boto3
from botocore.exceptions import ClientError
from six.moves.email_mime_multipart import MIMEMultipart
from six.moves.email_mime_text import MIMEText

import sure  # noqa

from moto import mock_ses


@mock_ses
def test_verify_email_identity():
    conn = boto3.client('ses', region_name='us-east-1')
    conn.verify_email_identity(EmailAddress="test@example.com")

    identities = conn.list_identities()
    address = identities['Identities'][0]
    address.should.equal('test@example.com')

@mock_ses
def test_verify_email_address():
    conn = boto3.client('ses', region_name='us-east-1')
    conn.verify_email_address(EmailAddress="test@example.com")
    email_addresses = conn.list_verified_email_addresses()
    email = email_addresses['VerifiedEmailAddresses'][0]
    email.should.equal('test@example.com')

@mock_ses
def test_domain_verify():
    conn = boto3.client('ses', region_name='us-east-1')

    conn.verify_domain_dkim(Domain="domain1.com")
    conn.verify_domain_identity(Domain="domain2.com")

    identities = conn.list_identities()
    domains = list(identities['Identities'])
    domains.should.equal(['domain1.com', 'domain2.com'])


@mock_ses
def test_delete_identity():
    conn = boto3.client('ses', region_name='us-east-1')
    conn.verify_email_identity(EmailAddress="test@example.com")

    conn.list_identities()['Identities'].should.have.length_of(1)
    conn.delete_identity(Identity="test@example.com")
    conn.list_identities()['Identities'].should.have.length_of(0)


@mock_ses
def test_send_email():
    conn = boto3.client('ses', region_name='us-east-1')

    kwargs = dict(
        Source="test@example.com",
        Destination={
            "ToAddresses": ["test_to@example.com"],
            "CcAddresses": ["test_cc@example.com"],
            "BccAddresses": ["test_bcc@example.com"],
        },
        Message={
            "Subject": {"Data": "test subject"},
            "Body": {"Text": {"Data": "test body"}}
        }
    )
    conn.send_email.when.called_with(**kwargs).should.throw(ClientError)

    conn.verify_domain_identity(Domain='example.com')
    conn.send_email(**kwargs)

    too_many_addresses = list('to%s@example.com' % i for i in range(51))
    conn.send_email.when.called_with(
        **dict(kwargs, Destination={'ToAddresses': too_many_addresses})
    ).should.throw(ClientError)

    send_quota = conn.get_send_quota()
    sent_count = int(send_quota['SentLast24Hours'])
    sent_count.should.equal(3)


@mock_ses
def test_send_html_email():
    conn = boto3.client('ses', region_name='us-east-1')

    kwargs = dict(
        Source="test@example.com",
        Destination={
            "ToAddresses": ["test_to@example.com"]
        },
        Message={
            "Subject": {"Data": "test subject"},
            "Body": {"Html": {"Data": "test body"}}
        }
    )

    conn.send_email.when.called_with(**kwargs).should.throw(ClientError)

    conn.verify_email_identity(EmailAddress="test@example.com")
    conn.send_email(**kwargs)

    send_quota = conn.get_send_quota()
    sent_count = int(send_quota['SentLast24Hours'])
    sent_count.should.equal(1)


@mock_ses
def test_send_raw_email():
    conn = boto3.client('ses', region_name='us-east-1')

    message = MIMEMultipart()
    message['Subject'] = 'Test'
    message['From'] = 'test@example.com'
    message['To'] = 'to@example.com, foo@example.com'

    # Message body
    part = MIMEText('test file attached')
    message.attach(part)

    # Attachment
    part = MIMEText('contents of test file here')
    part.add_header('Content-Disposition', 'attachment; filename=test.txt')
    message.attach(part)

    kwargs = dict(
        Source=message['From'],
        RawMessage={'Data': message.as_string()},
    )

    conn.send_raw_email.when.called_with(**kwargs).should.throw(ClientError)

    conn.verify_email_identity(EmailAddress="test@example.com")
    conn.send_raw_email(**kwargs)

    send_quota = conn.get_send_quota()
    sent_count = int(send_quota['SentLast24Hours'])
    sent_count.should.equal(2)



from __future__ import unicode_literals
from moto.core.responses import BaseResponse


class VPCPeeringConnections(BaseResponse):

    def create_vpc_peering_connection(self):
        vpc = self.ec2_backend.get_vpc(self._get_param('VpcId'))
        peer_vpc = self.ec2_backend.get_vpc(self._get_param('PeerVpcId'))
        vpc_pcx = self.ec2_backend.create_vpc_peering_connection(vpc, peer_vpc)
        template = self.response_template(
            CREATE_VPC_PEERING_CONNECTION_RESPONSE)
        return template.render(vpc_pcx=vpc_pcx)

    def delete_vpc_peering_connection(self):
        vpc_pcx_id = self._get_param('VpcPeeringConnectionId')
        vpc_pcx = self.ec2_backend.delete_vpc_peering_connection(vpc_pcx_id)
        template = self.response_template(
            DELETE_VPC_PEERING_CONNECTION_RESPONSE)
        return template.render(vpc_pcx=vpc_pcx)

    def describe_vpc_peering_connections(self):
        vpc_pcxs = self.ec2_backend.get_all_vpc_peering_connections()
        template = self.response_template(
            DESCRIBE_VPC_PEERING_CONNECTIONS_RESPONSE)
        return template.render(vpc_pcxs=vpc_pcxs)

    def accept_vpc_peering_connection(self):
        vpc_pcx_id = self._get_param('VpcPeeringConnectionId')
        vpc_pcx = self.ec2_backend.accept_vpc_peering_connection(vpc_pcx_id)
        template = self.response_template(
            ACCEPT_VPC_PEERING_CONNECTION_RESPONSE)
        return template.render(vpc_pcx=vpc_pcx)

    def reject_vpc_peering_connection(self):
        vpc_pcx_id = self._get_param('VpcPeeringConnectionId')
        self.ec2_backend.reject_vpc_peering_connection(vpc_pcx_id)
        template = self.response_template(
            REJECT_VPC_PEERING_CONNECTION_RESPONSE)
        return template.render()


CREATE_VPC_PEERING_CONNECTION_RESPONSE = """
<CreateVpcPeeringConnectionResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>7a62c49f-347e-4fc4-9331-6e8eEXAMPLE</requestId>
  <vpcPeeringConnection>
    <vpcPeeringConnectionId>{{ vpc_pcx.id }}</vpcPeeringConnectionId>
    <requesterVpcInfo>
      <ownerId>777788889999</ownerId>
      <vpcId>{{ vpc_pcx.vpc.id }}</vpcId>
      <cidrBlock>{{ vpc_pcx.vpc.cidr_block }}</cidrBlock>
    </requesterVpcInfo>
    <accepterVpcInfo>
      <ownerId>123456789012</ownerId>
      <vpcId>{{ vpc_pcx.peer_vpc.id }}</vpcId>
    </accepterVpcInfo>
    <status>
      <code>initiating-request</code>
      <message>Initiating request to {accepter ID}.</message>
    </status>
    <expirationTime>2014-02-18T14:37:25.000Z</expirationTime>
    <tagSet/>
  </vpcPeeringConnection>
</CreateVpcPeeringConnectionResponse>
"""

DESCRIBE_VPC_PEERING_CONNECTIONS_RESPONSE = """
<DescribeVpcPeeringConnectionsResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>7a62c49f-347e-4fc4-9331-6e8eEXAMPLE</requestId>
  <vpcPeeringConnectionSet>
    {% for vpc_pcx in vpc_pcxs %}
      <item>
        <vpcPeeringConnectionId>{{ vpc_pcx.id }}</vpcPeeringConnectionId>
        <requesterVpcInfo>
          <ownerId>777788889999</ownerId>
          <vpcId>{{ vpc_pcx.vpc.id }}</vpcId>
          <cidrBlock>{{ vpc_pcx.vpc.cidr_block }}</cidrBlock>
        </requesterVpcInfo>
        <accepterVpcInfo>
          <ownerId>123456789012</ownerId>
          <vpcId>{{ vpc_pcx.peer_vpc.id }}</vpcId>
        </accepterVpcInfo>
        <status>
          <code>{{ vpc_pcx._status.code }}</code>
          <message>{{ vpc_pcx._status.message }}</message>
        </status>
        <expirationTime>2014-02-17T16:00:50.000Z</expirationTime>
        <tagSet/>
      </item>
    {% endfor %}
  </vpcPeeringConnectionSet>
</DescribeVpcPeeringConnectionsResponse>
"""

DELETE_VPC_PEERING_CONNECTION_RESPONSE = """
<DeleteVpcPeeringConnectionResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>7a62c49f-347e-4fc4-9331-6e8eEXAMPLE</requestId>
  <return>true</return>
</DeleteVpcPeeringConnectionResponse>
"""

ACCEPT_VPC_PEERING_CONNECTION_RESPONSE = """
<AcceptVpcPeeringConnectionResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>7a62c49f-347e-4fc4-9331-6e8eEXAMPLE</requestId>
  <vpcPeeringConnection>
    <vpcPeeringConnectionId>{{ vpc_pcx.id }}</vpcPeeringConnectionId>
    <requesterVpcInfo>
      <ownerId>123456789012</ownerId>
      <vpcId>{{ vpc_pcx.vpc.id }}</vpcId>
      <cidrBlock>{{ vpc_pcx.vpc.cidr_block }}</cidrBlock>
    </requesterVpcInfo>
    <accepterVpcInfo>
      <ownerId>777788889999</ownerId>
      <vpcId>{{ vpc_pcx.peer_vpc.id }}</vpcId>
      <cidrBlock>{{ vpc_pcx.peer_vpc.cidr_block }}</cidrBlock>
    </accepterVpcInfo>
    <status>
      <code>{{ vpc_pcx._status.code }}</code>
      <message>{{ vpc_pcx._status.message }}</message>
    </status>
    <tagSet/>
  </vpcPeeringConnection>
</AcceptVpcPeeringConnectionResponse>
"""

REJECT_VPC_PEERING_CONNECTION_RESPONSE = """
<RejectVpcPeeringConnectionResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>7a62c49f-347e-4fc4-9331-6e8eEXAMPLE</requestId>
  <return>true</return>
</RejectVpcPeeringConnectionResponse>
"""



import boto3
import sure  # noqa

from moto import mock_logs, settings

_logs_region = 'us-east-1' if settings.TEST_SERVER_MODE else 'us-west-2'


@mock_logs
def test_log_group_create():
    conn = boto3.client('logs', 'us-west-2')
    log_group_name = 'dummy'
    response = conn.create_log_group(logGroupName=log_group_name)
    response = conn.delete_log_group(logGroupName=log_group_name)



from __future__ import unicode_literals
from .responses import InstanceMetadataResponse

url_bases = [
    "http://169.254.169.254"
]

instance_metadata = InstanceMetadataResponse()

url_paths = {
    '{0}/(?P<path>.+)': instance_metadata.metadata_response,
}



import boto
from boto.ec2.cloudwatch.alarm import MetricAlarm
import sure  # noqa

from moto import mock_cloudwatch_deprecated


def alarm_fixture(name="tester", action=None):
    action = action or ['arn:alarm']
    return MetricAlarm(
        name=name,
        namespace="{0}_namespace".format(name),
        metric="{0}_metric".format(name),
        comparison='>=',
        threshold=2.0,
        period=60,
        evaluation_periods=5,
        statistic='Average',
        description='A test',
        dimensions={'InstanceId': ['i-0123456,i-0123457']},
        alarm_actions=action,
        ok_actions=['arn:ok'],
        insufficient_data_actions=['arn:insufficient'],
        unit='Seconds',
    )


@mock_cloudwatch_deprecated
def test_create_alarm():
    conn = boto.connect_cloudwatch()

    alarm = alarm_fixture()
    conn.create_alarm(alarm)

    alarms = conn.describe_alarms()
    alarms.should.have.length_of(1)
    alarm = alarms[0]
    alarm.name.should.equal('tester')
    alarm.namespace.should.equal('tester_namespace')
    alarm.metric.should.equal('tester_metric')
    alarm.comparison.should.equal('>=')
    alarm.threshold.should.equal(2.0)
    alarm.period.should.equal(60)
    alarm.evaluation_periods.should.equal(5)
    alarm.statistic.should.equal('Average')
    alarm.description.should.equal('A test')
    dict(alarm.dimensions).should.equal(
        {'InstanceId': ['i-0123456,i-0123457']})
    list(alarm.alarm_actions).should.equal(['arn:alarm'])
    list(alarm.ok_actions).should.equal(['arn:ok'])
    list(alarm.insufficient_data_actions).should.equal(['arn:insufficient'])
    alarm.unit.should.equal('Seconds')


@mock_cloudwatch_deprecated
def test_delete_alarm():
    conn = boto.connect_cloudwatch()

    alarms = conn.describe_alarms()
    alarms.should.have.length_of(0)

    alarm = alarm_fixture()
    conn.create_alarm(alarm)

    alarms = conn.describe_alarms()
    alarms.should.have.length_of(1)

    alarms[0].delete()

    alarms = conn.describe_alarms()
    alarms.should.have.length_of(0)


@mock_cloudwatch_deprecated
def test_put_metric_data():
    conn = boto.connect_cloudwatch()

    conn.put_metric_data(
        namespace='tester',
        name='metric',
        value=1.5,
        dimensions={'InstanceId': ['i-0123456,i-0123457']},
    )

    metrics = conn.list_metrics()
    metrics.should.have.length_of(1)
    metric = metrics[0]
    metric.namespace.should.equal('tester')
    metric.name.should.equal('metric')
    dict(metric.dimensions).should.equal(
        {'InstanceId': ['i-0123456,i-0123457']})


@mock_cloudwatch_deprecated
def test_describe_alarms():
    conn = boto.connect_cloudwatch()

    alarms = conn.describe_alarms()
    alarms.should.have.length_of(0)

    conn.create_alarm(alarm_fixture(name="nfoobar", action="afoobar"))
    conn.create_alarm(alarm_fixture(name="nfoobaz", action="afoobaz"))
    conn.create_alarm(alarm_fixture(name="nbarfoo", action="abarfoo"))
    conn.create_alarm(alarm_fixture(name="nbazfoo", action="abazfoo"))

    alarms = conn.describe_alarms()
    alarms.should.have.length_of(4)
    alarms = conn.describe_alarms(alarm_name_prefix="nfoo")
    alarms.should.have.length_of(2)
    alarms = conn.describe_alarms(
        alarm_names=["nfoobar", "nbarfoo", "nbazfoo"])
    alarms.should.have.length_of(3)
    alarms = conn.describe_alarms(action_prefix="afoo")
    alarms.should.have.length_of(2)

    for alarm in conn.describe_alarms():
        alarm.delete()

    alarms = conn.describe_alarms()
    alarms.should.have.length_of(0)


@mock_cloudwatch_deprecated
def test_describe_state_value_unimplemented():
    conn = boto.connect_cloudwatch()

    conn.describe_alarms()
    conn.describe_alarms.when.called_with(
        state_value="foo").should.throw(NotImplementedError)



import json


class AWSError(Exception):
    CODE = None
    STATUS = 400

    def __init__(self, message, code=None, status=None):
        self.message = message
        self.code = code if code is not None else self.CODE
        self.status = status if status is not None else self.STATUS

    def response(self):
        return json.dumps({'__type': self.code, 'message': self.message}), dict(status=self.status)


class InvalidRequestException(AWSError):
    CODE = 'InvalidRequestException'


class BadSegmentException(Exception):
    def __init__(self, seg_id=None, code=None, message=None):
        self.id = seg_id
        self.code = code
        self.message = message

    def __repr__(self):
        return '<BadSegment {0}>'.format('-'.join([self.id, self.code, self.message]))

    def to_dict(self):
        result = {}
        if self.id is not None:
            result['Id'] = self.id
        if self.code is not None:
            result['ErrorCode'] = self.code
        if self.message is not None:
            result['Message'] = self.message

        return result



from __future__ import unicode_literals
from .responses import ELBV2Response

url_bases = [
    "https?://elasticloadbalancing.(.+).amazonaws.com",
]

url_paths = {
    '{0}/$': ELBV2Response.dispatch,
}



from __future__ import unicode_literals
from .responses import AWSCertificateManagerResponse

url_bases = [
    "https?://acm.(.+).amazonaws.com",
]

url_paths = {
    '{0}/$': AWSCertificateManagerResponse.dispatch,
}



from __future__ import unicode_literals

from moto.core.responses import BaseResponse
from moto.ec2.utils import filters_from_querystring


def try_parse_int(value, default=None):
    try:
        return int(value)
    except (TypeError, ValueError):
        return default


class SecurityGroups(BaseResponse):

    def _process_rules_from_querystring(self):
        group_name_or_id = (self._get_param('GroupName') or
                            self._get_param('GroupId'))

        querytree = {}
        for key, value in self.querystring.items():
            key_splitted = key.split('.')
            key_splitted = [try_parse_int(e, e) for e in key_splitted]

            d = querytree
            for subkey in key_splitted[:-1]:
                if subkey not in d:
                    d[subkey] = {}
                d = d[subkey]
            d[key_splitted[-1]] = value

        ip_permissions = querytree.get('IpPermissions') or {}
        for ip_permission_idx in sorted(ip_permissions.keys()):
            ip_permission = ip_permissions[ip_permission_idx]

            ip_protocol = ip_permission.get('IpProtocol', [None])[0]
            from_port = ip_permission.get('FromPort', [None])[0]
            to_port = ip_permission.get('ToPort', [None])[0]

            ip_ranges = []
            ip_ranges_tree = ip_permission.get('IpRanges') or {}
            for ip_range_idx in sorted(ip_ranges_tree.keys()):
                ip_ranges.append(ip_ranges_tree[ip_range_idx]['CidrIp'][0])

            source_groups = []
            source_group_ids = []
            groups_tree = ip_permission.get('Groups') or {}
            for group_idx in sorted(groups_tree.keys()):
                group_dict = groups_tree[group_idx]
                if 'GroupId' in group_dict:
                    source_group_ids.append(group_dict['GroupId'][0])
                elif 'GroupName' in group_dict:
                    source_groups.append(group_dict['GroupName'][0])

            yield (group_name_or_id, ip_protocol, from_port, to_port, ip_ranges,
                   source_groups, source_group_ids)

    def authorize_security_group_egress(self):
        if self.is_not_dryrun('GrantSecurityGroupEgress'):
            for args in self._process_rules_from_querystring():
                self.ec2_backend.authorize_security_group_egress(*args)
            return AUTHORIZE_SECURITY_GROUP_EGRESS_RESPONSE

    def authorize_security_group_ingress(self):
        if self.is_not_dryrun('GrantSecurityGroupIngress'):
            for args in self._process_rules_from_querystring():
                self.ec2_backend.authorize_security_group_ingress(*args)
            return AUTHORIZE_SECURITY_GROUP_INGRESS_REPONSE

    def create_security_group(self):
        name = self._get_param('GroupName')
        description = self._get_param('GroupDescription')
        vpc_id = self._get_param('VpcId')

        if self.is_not_dryrun('CreateSecurityGroup'):
            group = self.ec2_backend.create_security_group(
                name, description, vpc_id=vpc_id)
            template = self.response_template(CREATE_SECURITY_GROUP_RESPONSE)
            return template.render(group=group)

    def delete_security_group(self):
        # TODO this should raise an error if there are instances in the group.
        # See
        # http://docs.aws.amazon.com/AWSEC2/latest/APIReference/ApiReference-query-DeleteSecurityGroup.html

        name = self._get_param('GroupName')
        sg_id = self._get_param('GroupId')

        if self.is_not_dryrun('DeleteSecurityGroup'):
            if name:
                self.ec2_backend.delete_security_group(name)
            elif sg_id:
                self.ec2_backend.delete_security_group(group_id=sg_id)

            return DELETE_GROUP_RESPONSE

    def describe_security_groups(self):
        groupnames = self._get_multi_param("GroupName")
        group_ids = self._get_multi_param("GroupId")
        filters = filters_from_querystring(self.querystring)

        groups = self.ec2_backend.describe_security_groups(
            group_ids=group_ids,
            groupnames=groupnames,
            filters=filters
        )

        template = self.response_template(DESCRIBE_SECURITY_GROUPS_RESPONSE)
        return template.render(groups=groups)

    def revoke_security_group_egress(self):
        if self.is_not_dryrun('RevokeSecurityGroupEgress'):
            for args in self._process_rules_from_querystring():
                success = self.ec2_backend.revoke_security_group_egress(*args)
                if not success:
                    return "Could not find a matching egress rule", dict(status=404)
            return REVOKE_SECURITY_GROUP_EGRESS_RESPONSE

    def revoke_security_group_ingress(self):
        if self.is_not_dryrun('RevokeSecurityGroupIngress'):
            for args in self._process_rules_from_querystring():
                self.ec2_backend.revoke_security_group_ingress(*args)
            return REVOKE_SECURITY_GROUP_INGRESS_REPONSE


CREATE_SECURITY_GROUP_RESPONSE = """<CreateSecurityGroupResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
   <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
   <return>true</return>
   <groupId>{{ group.id }}</groupId>
</CreateSecurityGroupResponse>"""

DELETE_GROUP_RESPONSE = """<DeleteSecurityGroupResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</DeleteSecurityGroupResponse>"""

DESCRIBE_SECURITY_GROUPS_RESPONSE = """<DescribeSecurityGroupsResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
   <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
   <securityGroupInfo>
      {% for group in groups %}
          <item>
             <ownerId>123456789012</ownerId>
             <groupId>{{ group.id }}</groupId>
             <groupName>{{ group.name }}</groupName>
             <groupDescription>{{ group.description }}</groupDescription>
             {% if group.vpc_id %}
             <vpcId>{{ group.vpc_id }}</vpcId>
             {% endif %}
             <ipPermissions>
               {% for rule in group.ingress_rules %}
                    <item>
                       <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
                       {% if rule.from_port %}
                       <fromPort>{{ rule.from_port }}</fromPort>
                       {% endif %}
                       {% if rule.to_port %}
                       <toPort>{{ rule.to_port }}</toPort>
                       {% endif %}
                       <groups>
                          {% for source_group in rule.source_groups %}
                              <item>
                                 <userId>123456789012</userId>
                                 <groupId>{{ source_group.id }}</groupId>
                                 <groupName>{{ source_group.name }}</groupName>
                              </item>
                          {% endfor %}
                       </groups>
                       <ipRanges>
                          {% for ip_range in rule.ip_ranges %}
                              <item>
                                 <cidrIp>{{ ip_range }}</cidrIp>
                              </item>
                          {% endfor %}
                       </ipRanges>
                    </item>
                {% endfor %}
             </ipPermissions>
             <ipPermissionsEgress>
               {% for rule in group.egress_rules %}
                    <item>
                       <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
                       <fromPort>{{ rule.from_port }}</fromPort>
                       <toPort>{{ rule.to_port }}</toPort>
                       <groups>
                          {% for source_group in rule.source_groups %}
                              <item>
                                 <userId>123456789012</userId>
                                 <groupId>{{ source_group.id }}</groupId>
                                 <groupName>{{ source_group.name }}</groupName>
                              </item>
                          {% endfor %}
                       </groups>
                       <ipRanges>
                          {% for ip_range in rule.ip_ranges %}
                              <item>
                                 <cidrIp>{{ ip_range }}</cidrIp>
                              </item>
                          {% endfor %}
                       </ipRanges>
                    </item>
               {% endfor %}
             </ipPermissionsEgress>
             <tagSet>
               {% for tag in group.get_tags() %}
                 <item>
                   <resourceId>{{ tag.resource_id }}</resourceId>
                   <resourceType>{{ tag.resource_type }}</resourceType>
                   <key>{{ tag.key }}</key>
                   <value>{{ tag.value }}</value>
                 </item>
               {% endfor %}
             </tagSet>
          </item>
      {% endfor %}
   </securityGroupInfo>
</DescribeSecurityGroupsResponse>"""

AUTHORIZE_SECURITY_GROUP_INGRESS_REPONSE = """<AuthorizeSecurityGroupIngressResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</AuthorizeSecurityGroupIngressResponse>"""

REVOKE_SECURITY_GROUP_INGRESS_REPONSE = """<RevokeSecurityGroupIngressResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</RevokeSecurityGroupIngressResponse>"""

AUTHORIZE_SECURITY_GROUP_EGRESS_RESPONSE = """
<AuthorizeSecurityGroupEgressResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
   <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
   <return>true</return>
</AuthorizeSecurityGroupEgressResponse>"""

REVOKE_SECURITY_GROUP_EGRESS_RESPONSE = """<RevokeSecurityGroupEgressResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</RevokeSecurityGroupEgressResponse>"""



from __future__ import unicode_literals

template = {
    "Description": "AWS CloudFormation Sample Template vpc_single_instance_in_subnet.template: Sample template showing how to create a VPC and add an EC2 instance with an Elastic IP address and a security group. **WARNING** This template creates an Amazon EC2 instance. You will be billed for the AWS resources used if you create a stack from this template.",
    "Parameters": {
        "SSHLocation": {
            "ConstraintDescription": "must be a valid IP CIDR range of the form x.x.x.x/x.",
            "Description": " The IP address range that can be used to SSH to the EC2 instances",
            "Default": "0.0.0.0/0",
            "MinLength": "9",
            "AllowedPattern": "(\\d{1,3})\\.(\\d{1,3})\\.(\\d{1,3})\\.(\\d{1,3})/(\\d{1,2})",
            "MaxLength": "18",
            "Type": "String"
        },
        "KeyName": {
            "Type": "String",
            "Description": "Name of an existing EC2 KeyPair to enable SSH access to the instance",
            "MinLength": "1",
            "AllowedPattern": "[\\x20-\\x7E]*",
            "MaxLength": "255",
            "ConstraintDescription": "can contain only ASCII characters."
        },
        "InstanceType": {
            "Default": "m1.small",
            "ConstraintDescription": "must be a valid EC2 instance type.",
            "Type": "String",
            "Description": "WebServer EC2 instance type",
            "AllowedValues": [
                "t1.micro",
                "m1.small",
                "m1.medium",
                "m1.large",
                "m1.xlarge",
                "m2.xlarge",
                "m2.2xlarge",
                "m2.4xlarge",
                "m3.xlarge",
                "m3.2xlarge",
                "c1.medium",
                "c1.xlarge",
                "cc1.4xlarge",
                "cc2.8xlarge",
                "cg1.4xlarge"
            ]
        }
    },
    "AWSTemplateFormatVersion": "2010-09-09",
    "Outputs": {
        "URL": {
            "Description": "Newly created application URL",
            "Value": {
                "Fn::Join": [
                    "",
                    [
                        "http://",
                        {
                            "Fn::GetAtt": [
                                "WebServerInstance",
                                "PublicIp"
                            ]
                        }
                    ]
                ]
            }
        }
    },
    "Resources": {
        "Subnet": {
            "Type": "AWS::EC2::Subnet",
            "Properties": {
                "VpcId": {
                    "Ref": "VPC"
                },
                "CidrBlock": "10.0.0.0/24",
                "Tags": [
                    {
                        "Value": {
                            "Ref": "AWS::StackId"
                        },
                        "Key": "Application"
                    }
                ]
            }
        },
        "WebServerWaitHandle": {
            "Type": "AWS::CloudFormation::WaitConditionHandle"
        },
        "Route": {
            "Type": "AWS::EC2::Route",
            "Properties": {
                "GatewayId": {
                    "Ref": "InternetGateway"
                },
                "DestinationCidrBlock": "0.0.0.0/0",
                "RouteTableId": {
                    "Ref": "RouteTable"
                }
            },
            "DependsOn": "AttachGateway"
        },
        "SubnetRouteTableAssociation": {
            "Type": "AWS::EC2::SubnetRouteTableAssociation",
            "Properties": {
                "SubnetId": {
                    "Ref": "Subnet"
                },
                "RouteTableId": {
                    "Ref": "RouteTable"
                }
            }
        },
        "InternetGateway": {
            "Type": "AWS::EC2::InternetGateway",
            "Properties": {
                "Tags": [
                    {
                        "Value": {
                            "Ref": "AWS::StackId"
                        },
                        "Key": "Application"
                    }
                ]
            }
        },
        "RouteTable": {
            "Type": "AWS::EC2::RouteTable",
            "Properties": {
                "VpcId": {
                    "Ref": "VPC"
                },
                "Tags": [
                    {
                        "Value": {
                            "Ref": "AWS::StackId"
                        },
                        "Key": "Application"
                    }
                ]
            }
        },
        "WebServerWaitCondition": {
            "Type": "AWS::CloudFormation::WaitCondition",
            "Properties": {
                "Handle": {
                    "Ref": "WebServerWaitHandle"
                },
                "Timeout": "300"
            },
            "DependsOn": "WebServerInstance"
        },
        "VPC": {
            "Type": "AWS::EC2::VPC",
            "Properties": {
                "CidrBlock": "10.0.0.0/16",
                "Tags": [
                    {
                        "Value": {
                            "Ref": "AWS::StackId"
                        },
                        "Key": "Application"
                    }
                ]
            }
        },
        "InstanceSecurityGroup": {
            "Type": "AWS::EC2::SecurityGroup",
            "Properties": {
                "SecurityGroupIngress": [
                    {
                        "ToPort": "22",
                        "IpProtocol": "tcp",
                        "CidrIp": {
                            "Ref": "SSHLocation"
                        },
                        "FromPort": "22"
                    },
                    {
                        "ToPort": "80",
                        "IpProtocol": "tcp",
                        "CidrIp": "0.0.0.0/0",
                        "FromPort": "80"
                    }
                ],
                "VpcId": {
                    "Ref": "VPC"
                },
                "GroupDescription": "Enable SSH access via port 22"
            }
        },
        "WebServerInstance": {
            "Type": "AWS::EC2::Instance",
            "Properties": {
                "UserData": {
                    "Fn::Base64": {
                        "Fn::Join": [
                            "",
                            [
                                "#!/bin/bash\n",
                                "yum update -y aws-cfn-bootstrap\n",
                                "# Helper function\n",
                                "function error_exit\n",
                                "{\n",
                                "  /opt/aws/bin/cfn-signal -e 1 -r \"$1\" '",
                                {
                                    "Ref": "WebServerWaitHandle"
                                },
                                "'\n",
                                "  exit 1\n",
                                "}\n",
                                "# Install the simple web page\n",
                                "/opt/aws/bin/cfn-init -s ",
                                {
                                    "Ref": "AWS::StackId"
                                },
                                " -r WebServerInstance ",
                                "         --region ",
                                {
                                    "Ref": "AWS::Region"
                                },
                                " || error_exit 'Failed to run cfn-init'\n",
                                "# Start up the cfn-hup daemon to listen for changes to the Web Server metadata\n",
                                "/opt/aws/bin/cfn-hup || error_exit 'Failed to start cfn-hup'\n",
                                "# All done so signal success\n",
                                "/opt/aws/bin/cfn-signal -e 0 -r \"WebServer setup complete\" '",
                                {
                                    "Ref": "WebServerWaitHandle"
                                },
                                "'\n"
                            ]
                        ]
                    }
                },
                "Tags": [
                    {
                        "Value": {
                            "Ref": "AWS::StackId"
                        },
                        "Key": "Application"
                    },
                    {
                        "Value": "Bar",
                        "Key": "Foo"
                    }
                ],
                "SecurityGroupIds": [
                    {
                        "Ref": "InstanceSecurityGroup"
                    }
                ],
                "KeyName": {
                    "Ref": "KeyName"
                },
                "SubnetId": {
                    "Ref": "Subnet"
                },
                "ImageId": {
                    "Fn::FindInMap": [
                        "RegionMap",
                        {
                            "Ref": "AWS::Region"
                        },
                        "AMI"
                    ]
                },
                "InstanceType": {
                    "Ref": "InstanceType"
                }
            },
            "Metadata": {
                "Comment": "Install a simple PHP application",
                "AWS::CloudFormation::Init": {
                    "config": {
                        "files": {
                            "/etc/cfn/cfn-hup.conf": {
                                "content": {
                                    "Fn::Join": [
                                        "",
                                        [
                                            "[main]\n",
                                            "stack=",
                                            {
                                                "Ref": "AWS::StackId"
                                            },
                                            "\n",
                                            "region=",
                                            {
                                                "Ref": "AWS::Region"
                                            },
                                            "\n"
                                        ]
                                    ]
                                },
                                "owner": "root",
                                "group": "root",
                                "mode": "000400"
                            },
                            "/etc/cfn/hooks.d/cfn-auto-reloader.conf": {
                                "content": {
                                    "Fn::Join": [
                                        "",
                                        [
                                            "[cfn-auto-reloader-hook]\n",
                                            "triggers=post.update\n",
                                            "path=Resources.WebServerInstance.Metadata.AWS::CloudFormation::Init\n",
                                            "action=/opt/aws/bin/cfn-init -s ",
                                            {
                                                "Ref": "AWS::StackId"
                                            },
                                            " -r WebServerInstance ",
                                            " --region     ",
                                            {
                                                "Ref": "AWS::Region"
                                            },
                                            "\n",
                                            "runas=root\n"
                                        ]
                                    ]
                                }
                            },
                            "/var/www/html/index.php": {
                                "content": {
                                    "Fn::Join": [
                                        "",
                                        [
                                            "<?php\n",
                                            "echo '<h1>AWS CloudFormation sample PHP application</h1>';\n",
                                            "?>\n"
                                        ]
                                    ]
                                },
                                "owner": "apache",
                                "group": "apache",
                                "mode": "000644"
                            }
                        },
                        "services": {
                            "sysvinit": {
                                "httpd": {
                                    "ensureRunning": "true",
                                    "enabled": "true"
                                },
                                "sendmail": {
                                    "ensureRunning": "false",
                                    "enabled": "false"
                                }
                            }
                        },
                        "packages": {
                            "yum": {
                                "httpd": [],
                                "php": []
                            }
                        }
                    }
                }
            }
        },
        "IPAddress": {
            "Type": "AWS::EC2::EIP",
            "Properties": {
                "InstanceId": {
                    "Ref": "WebServerInstance"
                },
                "Domain": "vpc"
            },
            "DependsOn": "AttachGateway"
        },
        "AttachGateway": {
            "Type": "AWS::EC2::VPCGatewayAttachment",
            "Properties": {
                "VpcId": {
                    "Ref": "VPC"
                },
                "InternetGatewayId": {
                    "Ref": "InternetGateway"
                }
            }
        }
    },
    "Mappings": {
        "RegionMap": {
            "ap-southeast-1": {
                "AMI": "ami-74dda626"
            },
            "ap-southeast-2": {
                "AMI": "ami-b3990e89"
            },
            "us-west-2": {
                "AMI": "ami-16fd7026"
            },
            "us-east-1": {
                "AMI": "ami-7f418316"
            },
            "ap-northeast-1": {
                "AMI": "ami-dcfa4edd"
            },
            "us-west-1": {
                "AMI": "ami-951945d0"
            },
            "eu-west-1": {
                "AMI": "ami-24506250"
            },
            "sa-east-1": {
                "AMI": "ami-3e3be423"
            }
        }
    }
}



from __future__ import unicode_literals

from collections import defaultdict

from moto.core import BaseBackend, BaseModel
from moto.ec2 import ec2_backends


class Parameter(BaseModel):
    def __init__(self, name, value, type, description, keyid):
        self.name = name
        self.type = type
        self.description = description
        self.keyid = keyid

        if self.type == 'SecureString':
            self.value = self.encrypt(value)
        else:
            self.value = value

    def encrypt(self, value):
        return 'kms:{}:'.format(self.keyid or 'default') + value

    def decrypt(self, value):
        if self.type != 'SecureString':
            return value

        prefix = 'kms:{}:'.format(self.keyid or 'default')
        if value.startswith(prefix):
            return value[len(prefix):]

    def response_object(self, decrypt=False):
        r = {
            'Name': self.name,
            'Type': self.type,
            'Value': self.decrypt(self.value) if decrypt else self.value
        }
        if self.keyid:
            r['KeyId'] = self.keyid
        return r


class SimpleSystemManagerBackend(BaseBackend):

    def __init__(self):
        self._parameters = {}
        self._resource_tags = defaultdict(lambda: defaultdict(dict))

    def delete_parameter(self, name):
        try:
            del self._parameters[name]
        except KeyError:
            pass

    def delete_parameters(self, names):
        result = []
        for name in names:
            try:
                del self._parameters[name]
                result.append(name)
            except KeyError:
                pass
        return result

    def get_all_parameters(self):
        result = []
        for k, _ in self._parameters.items():
            result.append(self._parameters[k])
        return result

    def get_parameters(self, names, with_decryption):
        result = []
        for name in names:
            if name in self._parameters:
                result.append(self._parameters[name])
        return result

    def get_parameter(self, name, with_decryption):
        if name in self._parameters:
            return self._parameters[name]
        return None

    def put_parameter(self, name, description, value, type, keyid, overwrite):
        if not overwrite and name in self._parameters:
            return
        self._parameters[name] = Parameter(
            name, value, type, description, keyid)

    def add_tags_to_resource(self, resource_type, resource_id, tags):
        for key, value in tags.items():
            self._resource_tags[resource_type][resource_id][key] = value

    def remove_tags_from_resource(self, resource_type, resource_id, keys):
        tags = self._resource_tags[resource_type][resource_id]
        for key in keys:
            if key in tags:
                del tags[key]

    def list_tags_for_resource(self, resource_type, resource_id):
        return self._resource_tags[resource_type][resource_id]


ssm_backends = {}
for region, ec2_backend in ec2_backends.items():
    ssm_backends[region] = SimpleSystemManagerBackend()



from __future__ import unicode_literals
import boto
import sure  # noqa

from moto import mock_ec2_deprecated


@mock_ec2_deprecated
def test_default_network_acl_created_with_vpc():
    conn = boto.connect_vpc('the_key', 'the secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    all_network_acls = conn.get_all_network_acls()
    all_network_acls.should.have.length_of(2)


@mock_ec2_deprecated
def test_network_acls():
    conn = boto.connect_vpc('the_key', 'the secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    network_acl = conn.create_network_acl(vpc.id)
    all_network_acls = conn.get_all_network_acls()
    all_network_acls.should.have.length_of(3)


@mock_ec2_deprecated
def test_new_subnet_associates_with_default_network_acl():
    conn = boto.connect_vpc('the_key', 'the secret')
    vpc = conn.get_all_vpcs()[0]

    subnet = conn.create_subnet(vpc.id, "10.0.0.0/18")
    all_network_acls = conn.get_all_network_acls()
    all_network_acls.should.have.length_of(1)

    acl = all_network_acls[0]
    acl.associations.should.have.length_of(4)
    [a.subnet_id for a in acl.associations].should.contain(subnet.id)


@mock_ec2_deprecated
def test_network_acl_entries():
    conn = boto.connect_vpc('the_key', 'the secret')
    vpc = conn.create_vpc("10.0.0.0/16")

    network_acl = conn.create_network_acl(vpc.id)

    network_acl_entry = conn.create_network_acl_entry(
        network_acl.id, 110, 6,
        'ALLOW', '0.0.0.0/0', False,
        port_range_from='443',
        port_range_to='443'
    )

    all_network_acls = conn.get_all_network_acls()
    all_network_acls.should.have.length_of(3)

    test_network_acl = next(na for na in all_network_acls
                            if na.id == network_acl.id)
    entries = test_network_acl.network_acl_entries
    entries.should.have.length_of(1)
    entries[0].rule_number.should.equal('110')
    entries[0].protocol.should.equal('6')
    entries[0].rule_action.should.equal('ALLOW')


@mock_ec2_deprecated
def test_delete_network_acl_entry():
    conn = boto.connect_vpc('the_key', 'the secret')
    vpc = conn.create_vpc("10.0.0.0/16")

    network_acl = conn.create_network_acl(vpc.id)

    conn.create_network_acl_entry(
        network_acl.id, 110, 6,
        'ALLOW', '0.0.0.0/0', False,
        port_range_from='443',
        port_range_to='443'
    )
    conn.delete_network_acl_entry(
        network_acl.id, 110, False
    )

    all_network_acls = conn.get_all_network_acls()

    test_network_acl = next(na for na in all_network_acls
                            if na.id == network_acl.id)
    entries = test_network_acl.network_acl_entries
    entries.should.have.length_of(0)


@mock_ec2_deprecated
def test_replace_network_acl_entry():
    conn = boto.connect_vpc('the_key', 'the secret')
    vpc = conn.create_vpc("10.0.0.0/16")

    network_acl = conn.create_network_acl(vpc.id)

    conn.create_network_acl_entry(
        network_acl.id, 110, 6,
        'ALLOW', '0.0.0.0/0', False,
        port_range_from='443',
        port_range_to='443'
    )
    conn.replace_network_acl_entry(
        network_acl.id, 110, -1,
        'DENY', '0.0.0.0/0', False,
        port_range_from='22',
        port_range_to='22'
    )

    all_network_acls = conn.get_all_network_acls()

    test_network_acl = next(na for na in all_network_acls
                            if na.id == network_acl.id)
    entries = test_network_acl.network_acl_entries
    entries.should.have.length_of(1)
    entries[0].rule_number.should.equal('110')
    entries[0].protocol.should.equal('-1')
    entries[0].rule_action.should.equal('DENY')

@mock_ec2_deprecated
def test_associate_new_network_acl_with_subnet():
    conn = boto.connect_vpc('the_key', 'the secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    subnet = conn.create_subnet(vpc.id, "10.0.0.0/18")
    network_acl = conn.create_network_acl(vpc.id)

    conn.associate_network_acl(network_acl.id, subnet.id)

    all_network_acls = conn.get_all_network_acls()
    all_network_acls.should.have.length_of(3)

    test_network_acl = next(na for na in all_network_acls
                            if na.id == network_acl.id)

    test_network_acl.associations.should.have.length_of(1)
    test_network_acl.associations[0].subnet_id.should.equal(subnet.id)


@mock_ec2_deprecated
def test_delete_network_acl():
    conn = boto.connect_vpc('the_key', 'the secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    subnet = conn.create_subnet(vpc.id, "10.0.0.0/18")
    network_acl = conn.create_network_acl(vpc.id)

    all_network_acls = conn.get_all_network_acls()
    all_network_acls.should.have.length_of(3)

    any(acl.id == network_acl.id for acl in all_network_acls).should.be.ok

    conn.delete_network_acl(network_acl.id)

    updated_network_acls = conn.get_all_network_acls()
    updated_network_acls.should.have.length_of(2)

    any(acl.id == network_acl.id for acl in updated_network_acls).shouldnt.be.ok


@mock_ec2_deprecated
def test_network_acl_tagging():
    conn = boto.connect_vpc('the_key', 'the secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    network_acl = conn.create_network_acl(vpc.id)

    network_acl.add_tag("a key", "some value")

    tag = conn.get_all_tags()[0]
    tag.name.should.equal("a key")
    tag.value.should.equal("some value")

    all_network_acls = conn.get_all_network_acls()
    test_network_acl = next(na for na in all_network_acls
                            if na.id == network_acl.id)
    test_network_acl.tags.should.have.length_of(1)
    test_network_acl.tags["a key"].should.equal("some value")



from __future__ import unicode_literals
# Ensure 'assert_raises' context manager support for Python 2.6
import tests.backport_assert_raises
from nose.tools import assert_raises

import boto3
from botocore.exceptions import ClientError
import boto
import boto.cloudformation
import boto.ec2
from boto.exception import EC2ResponseError
import sure  # noqa

from moto import mock_ec2, mock_cloudformation_deprecated, mock_ec2_deprecated
from tests.helpers import requires_boto_gte
from tests.test_cloudformation.fixtures import vpc_eni
import json


@mock_ec2_deprecated
def test_elastic_network_interfaces():
    conn = boto.connect_vpc('the_key', 'the_secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    subnet = conn.create_subnet(vpc.id, "10.0.0.0/18")

    with assert_raises(EC2ResponseError) as ex:
        eni = conn.create_network_interface(subnet.id, dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the CreateNetworkInterface operation: Request would have succeeded, but DryRun flag is set')

    eni = conn.create_network_interface(subnet.id)

    all_enis = conn.get_all_network_interfaces()
    all_enis.should.have.length_of(1)
    eni = all_enis[0]
    eni.groups.should.have.length_of(0)
    eni.private_ip_addresses.should.have.length_of(0)

    with assert_raises(EC2ResponseError) as ex:
        conn.delete_network_interface(eni.id, dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the DeleteNetworkInterface operation: Request would have succeeded, but DryRun flag is set')

    conn.delete_network_interface(eni.id)

    all_enis = conn.get_all_network_interfaces()
    all_enis.should.have.length_of(0)

    with assert_raises(EC2ResponseError) as cm:
        conn.delete_network_interface(eni.id)
    cm.exception.error_code.should.equal('InvalidNetworkInterfaceID.NotFound')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none


@mock_ec2_deprecated
def test_elastic_network_interfaces_subnet_validation():
    conn = boto.connect_vpc('the_key', 'the_secret')

    with assert_raises(EC2ResponseError) as cm:
        conn.create_network_interface("subnet-abcd1234")
    cm.exception.error_code.should.equal('InvalidSubnetID.NotFound')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none


@mock_ec2_deprecated
def test_elastic_network_interfaces_with_private_ip():
    conn = boto.connect_vpc('the_key', 'the_secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    subnet = conn.create_subnet(vpc.id, "10.0.0.0/18")
    private_ip = "54.0.0.1"
    eni = conn.create_network_interface(subnet.id, private_ip)

    all_enis = conn.get_all_network_interfaces()
    all_enis.should.have.length_of(1)

    eni = all_enis[0]
    eni.groups.should.have.length_of(0)

    eni.private_ip_addresses.should.have.length_of(1)
    eni.private_ip_addresses[0].private_ip_address.should.equal(private_ip)


@mock_ec2_deprecated
def test_elastic_network_interfaces_with_groups():
    conn = boto.connect_vpc('the_key', 'the_secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    subnet = conn.create_subnet(vpc.id, "10.0.0.0/18")
    security_group1 = conn.create_security_group(
        'test security group #1', 'this is a test security group')
    security_group2 = conn.create_security_group(
        'test security group #2', 'this is a test security group')
    conn.create_network_interface(
        subnet.id, groups=[security_group1.id, security_group2.id])

    all_enis = conn.get_all_network_interfaces()
    all_enis.should.have.length_of(1)

    eni = all_enis[0]
    eni.groups.should.have.length_of(2)
    set([group.id for group in eni.groups]).should.equal(
        set([security_group1.id, security_group2.id]))


@requires_boto_gte("2.12.0")
@mock_ec2_deprecated
def test_elastic_network_interfaces_modify_attribute():
    conn = boto.connect_vpc('the_key', 'the_secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    subnet = conn.create_subnet(vpc.id, "10.0.0.0/18")
    security_group1 = conn.create_security_group(
        'test security group #1', 'this is a test security group')
    security_group2 = conn.create_security_group(
        'test security group #2', 'this is a test security group')
    conn.create_network_interface(subnet.id, groups=[security_group1.id])

    all_enis = conn.get_all_network_interfaces()
    all_enis.should.have.length_of(1)

    eni = all_enis[0]
    eni.groups.should.have.length_of(1)
    eni.groups[0].id.should.equal(security_group1.id)

    with assert_raises(EC2ResponseError) as ex:
        conn.modify_network_interface_attribute(
            eni.id, 'groupset', [security_group2.id], dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the ModifyNetworkInterface operation: Request would have succeeded, but DryRun flag is set')

    conn.modify_network_interface_attribute(
        eni.id, 'groupset', [security_group2.id])

    all_enis = conn.get_all_network_interfaces()
    all_enis.should.have.length_of(1)

    eni = all_enis[0]
    eni.groups.should.have.length_of(1)
    eni.groups[0].id.should.equal(security_group2.id)


@mock_ec2_deprecated
def test_elastic_network_interfaces_filtering():
    conn = boto.connect_vpc('the_key', 'the_secret')
    vpc = conn.create_vpc("10.0.0.0/16")
    subnet = conn.create_subnet(vpc.id, "10.0.0.0/18")

    security_group1 = conn.create_security_group(
        'test security group #1', 'this is a test security group')
    security_group2 = conn.create_security_group(
        'test security group #2', 'this is a test security group')

    eni1 = conn.create_network_interface(
        subnet.id, groups=[security_group1.id, security_group2.id])
    eni2 = conn.create_network_interface(
        subnet.id, groups=[security_group1.id])
    eni3 = conn.create_network_interface(subnet.id)

    all_enis = conn.get_all_network_interfaces()
    all_enis.should.have.length_of(3)

    # Filter by NetworkInterfaceId
    enis_by_id = conn.get_all_network_interfaces([eni1.id])
    enis_by_id.should.have.length_of(1)
    set([eni.id for eni in enis_by_id]).should.equal(set([eni1.id]))

    # Filter by ENI ID
    enis_by_id = conn.get_all_network_interfaces(
        filters={'network-interface-id': eni1.id})
    enis_by_id.should.have.length_of(1)
    set([eni.id for eni in enis_by_id]).should.equal(set([eni1.id]))

    # Filter by Security Group
    enis_by_group = conn.get_all_network_interfaces(
        filters={'group-id': security_group1.id})
    enis_by_group.should.have.length_of(2)
    set([eni.id for eni in enis_by_group]).should.equal(set([eni1.id, eni2.id]))

    # Filter by ENI ID and Security Group
    enis_by_group = conn.get_all_network_interfaces(
        filters={'network-interface-id': eni1.id, 'group-id': security_group1.id})
    enis_by_group.should.have.length_of(1)
    set([eni.id for eni in enis_by_group]).should.equal(set([eni1.id]))

    # Unsupported filter
    conn.get_all_network_interfaces.when.called_with(
        filters={'not-implemented-filter': 'foobar'}).should.throw(NotImplementedError)


@mock_ec2
def test_elastic_network_interfaces_get_by_tag_name():
    ec2 = boto3.resource('ec2', region_name='us-west-2')
    ec2_client = boto3.client('ec2', region_name='us-west-2')

    vpc = ec2.create_vpc(CidrBlock='10.0.0.0/16')
    subnet = ec2.create_subnet(
        VpcId=vpc.id, CidrBlock='10.0.0.0/24', AvailabilityZone='us-west-2a')

    eni1 = ec2.create_network_interface(
        SubnetId=subnet.id, PrivateIpAddress='10.0.10.5')

    with assert_raises(ClientError) as ex:
        eni1.create_tags(Tags=[{'Key': 'Name', 'Value': 'eni1'}], DryRun=True)
    ex.exception.response['Error']['Code'].should.equal('DryRunOperation')
    ex.exception.response['ResponseMetadata'][
        'HTTPStatusCode'].should.equal(400)
    ex.exception.response['Error']['Message'].should.equal(
        'An error occurred (DryRunOperation) when calling the CreateTags operation: Request would have succeeded, but DryRun flag is set')

    eni1.create_tags(Tags=[{'Key': 'Name', 'Value': 'eni1'}])

    # The status of the new interface should be 'available'
    waiter = ec2_client.get_waiter('network_interface_available')
    waiter.wait(NetworkInterfaceIds=[eni1.id])

    filters = [{'Name': 'tag:Name', 'Values': ['eni1']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(1)

    filters = [{'Name': 'tag:Name', 'Values': ['wrong-name']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(0)


@mock_ec2
def test_elastic_network_interfaces_get_by_availability_zone():
    ec2 = boto3.resource('ec2', region_name='us-west-2')
    ec2_client = boto3.client('ec2', region_name='us-west-2')

    vpc = ec2.create_vpc(CidrBlock='10.0.0.0/16')
    subnet1 = ec2.create_subnet(
        VpcId=vpc.id, CidrBlock='10.0.0.0/24', AvailabilityZone='us-west-2a')

    subnet2 = ec2.create_subnet(
        VpcId=vpc.id, CidrBlock='10.0.1.0/24', AvailabilityZone='us-west-2b')

    eni1 = ec2.create_network_interface(
        SubnetId=subnet1.id, PrivateIpAddress='10.0.0.15')

    eni2 = ec2.create_network_interface(
        SubnetId=subnet2.id, PrivateIpAddress='10.0.1.15')

    # The status of the new interface should be 'available'
    waiter = ec2_client.get_waiter('network_interface_available')
    waiter.wait(NetworkInterfaceIds=[eni1.id, eni2.id])

    filters = [{'Name': 'availability-zone', 'Values': ['us-west-2a']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(1)

    filters = [{'Name': 'availability-zone', 'Values': ['us-west-2c']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(0)


@mock_ec2
def test_elastic_network_interfaces_get_by_private_ip():
    ec2 = boto3.resource('ec2', region_name='us-west-2')
    ec2_client = boto3.client('ec2', region_name='us-west-2')

    vpc = ec2.create_vpc(CidrBlock='10.0.0.0/16')
    subnet = ec2.create_subnet(
        VpcId=vpc.id, CidrBlock='10.0.0.0/24', AvailabilityZone='us-west-2a')

    eni1 = ec2.create_network_interface(
        SubnetId=subnet.id, PrivateIpAddress='10.0.10.5')

    # The status of the new interface should be 'available'
    waiter = ec2_client.get_waiter('network_interface_available')
    waiter.wait(NetworkInterfaceIds=[eni1.id])

    filters = [{'Name': 'private-ip-address', 'Values': ['10.0.10.5']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(1)

    filters = [{'Name': 'private-ip-address', 'Values': ['10.0.10.10']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(0)

    filters = [{'Name': 'addresses.private-ip-address', 'Values': ['10.0.10.5']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(1)

    filters = [{'Name': 'addresses.private-ip-address', 'Values': ['10.0.10.10']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(0)


@mock_ec2
def test_elastic_network_interfaces_get_by_vpc_id():
    ec2 = boto3.resource('ec2', region_name='us-west-2')
    ec2_client = boto3.client('ec2', region_name='us-west-2')

    vpc = ec2.create_vpc(CidrBlock='10.0.0.0/16')
    subnet = ec2.create_subnet(
        VpcId=vpc.id, CidrBlock='10.0.0.0/24', AvailabilityZone='us-west-2a')

    eni1 = ec2.create_network_interface(
        SubnetId=subnet.id, PrivateIpAddress='10.0.10.5')

    # The status of the new interface should be 'available'
    waiter = ec2_client.get_waiter('network_interface_available')
    waiter.wait(NetworkInterfaceIds=[eni1.id])

    filters = [{'Name': 'vpc-id', 'Values': [subnet.vpc_id]}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(1)

    filters = [{'Name': 'vpc-id', 'Values': ['vpc-aaaa1111']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(0)


@mock_ec2
def test_elastic_network_interfaces_get_by_subnet_id():
    ec2 = boto3.resource('ec2', region_name='us-west-2')
    ec2_client = boto3.client('ec2', region_name='us-west-2')

    vpc = ec2.create_vpc(CidrBlock='10.0.0.0/16')
    subnet = ec2.create_subnet(
        VpcId=vpc.id, CidrBlock='10.0.0.0/24', AvailabilityZone='us-west-2a')

    eni1 = ec2.create_network_interface(
        SubnetId=subnet.id, PrivateIpAddress='10.0.10.5')

    # The status of the new interface should be 'available'
    waiter = ec2_client.get_waiter('network_interface_available')
    waiter.wait(NetworkInterfaceIds=[eni1.id])

    filters = [{'Name': 'subnet-id', 'Values': [subnet.id]}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(1)

    filters = [{'Name': 'subnet-id', 'Values': ['subnet-aaaa1111']}]
    enis = list(ec2.network_interfaces.filter(Filters=filters))
    enis.should.have.length_of(0)


@mock_ec2_deprecated
@mock_cloudformation_deprecated
def test_elastic_network_interfaces_cloudformation():
    template = vpc_eni.template
    template_json = json.dumps(template)
    conn = boto.cloudformation.connect_to_region("us-west-1")
    conn.create_stack(
        "test_stack",
        template_body=template_json,
    )
    ec2_conn = boto.ec2.connect_to_region("us-west-1")
    eni = ec2_conn.get_all_network_interfaces()[0]

    stack = conn.describe_stacks()[0]
    resources = stack.describe_resources()
    cfn_eni = [resource for resource in resources if resource.resource_type ==
               'AWS::EC2::NetworkInterface'][0]
    cfn_eni.physical_resource_id.should.equal(eni.id)



from __future__ import unicode_literals

template = {
    "Resources": {
        "Ec2Instance": {
            "Type": "AWS::EC2::Instance",
            "Properties": {
                "ImageId": "ami-1234abcd",
                "PrivateIpAddress": "10.0.0.25",
            }
        },

        "HostedZone": {
            "Type": "AWS::Route53::HostedZone",
            "Properties": {
                "Name": "my_zone"
            }
        },

        "myDNSRecord": {
            "Type": "AWS::Route53::RecordSet",
            "Properties": {
                "HostedZoneName": {"Ref": "HostedZone"},
                "Comment": "DNS name for my instance.",
                "Name": {
                    "Fn::Join": ["", [
                        {"Ref": "Ec2Instance"}, ".",
                        {"Ref": "AWS::Region"}, ".",
                        {"Ref": "HostedZone"}, "."
                    ]]
                },
                "Type": "A",
                "TTL": "900",
                "ResourceRecords": [
                    {"Fn::GetAtt": ["Ec2Instance", "PrivateIp"]}
                ]
            }
        }
    },
}



from __future__ import unicode_literals

import boto3
import botocore.exceptions
import sure   # noqa

from moto import mock_ssm


@mock_ssm
def test_delete_parameter():
    client = boto3.client('ssm', region_name='us-east-1')

    client.put_parameter(
        Name='test',
        Description='A test parameter',
        Value='value',
        Type='String')

    response = client.get_parameters(Names=['test'])
    len(response['Parameters']).should.equal(1)

    client.delete_parameter(Name='test')

    response = client.get_parameters(Names=['test'])
    len(response['Parameters']).should.equal(0)


@mock_ssm
def test_delete_parameters():
    client = boto3.client('ssm', region_name='us-east-1')

    client.put_parameter(
        Name='test',
        Description='A test parameter',
        Value='value',
        Type='String')

    response = client.get_parameters(Names=['test'])
    len(response['Parameters']).should.equal(1)

    result = client.delete_parameters(Names=['test', 'invalid'])
    len(result['DeletedParameters']).should.equal(1)
    len(result['InvalidParameters']).should.equal(1)

    response = client.get_parameters(Names=['test'])
    len(response['Parameters']).should.equal(0)


@mock_ssm
def test_put_parameter():
    client = boto3.client('ssm', region_name='us-east-1')

    client.put_parameter(
        Name='test',
        Description='A test parameter',
        Value='value',
        Type='String')

    response = client.get_parameters(
        Names=[
            'test'
        ],
        WithDecryption=False)

    len(response['Parameters']).should.equal(1)
    response['Parameters'][0]['Name'].should.equal('test')
    response['Parameters'][0]['Value'].should.equal('value')
    response['Parameters'][0]['Type'].should.equal('String')


@mock_ssm
def test_get_parameter():
    client = boto3.client('ssm', region_name='us-east-1')

    client.put_parameter(
        Name='test',
        Description='A test parameter',
        Value='value',
        Type='String')

    response = client.get_parameter(
        Name='test',
        WithDecryption=False)

    response['Parameter']['Name'].should.equal('test')
    response['Parameter']['Value'].should.equal('value')
    response['Parameter']['Type'].should.equal('String')


@mock_ssm
def test_get_nonexistant_parameter():
    client = boto3.client('ssm', region_name='us-east-1')

    try:
        client.get_parameter(
            Name='test_noexist',
            WithDecryption=False)
        raise RuntimeError('Should of failed')
    except botocore.exceptions.ClientError as err:
        err.operation_name.should.equal('GetParameter')
        err.response['Error']['Message'].should.equal('Parameter test_noexist not found.')


@mock_ssm
def test_describe_parameters():
    client = boto3.client('ssm', region_name='us-east-1')

    client.put_parameter(
        Name='test',
        Description='A test parameter',
        Value='value',
        Type='String')

    response = client.describe_parameters()

    len(response['Parameters']).should.equal(1)
    response['Parameters'][0]['Name'].should.equal('test')
    response['Parameters'][0]['Type'].should.equal('String')


@mock_ssm
def test_describe_parameters_paging():
    client = boto3.client('ssm', region_name='us-east-1')

    for i in range(50):
        client.put_parameter(
            Name="param-%d" % i,
            Value="value-%d" % i,
            Type="String"
        )

    response = client.describe_parameters()
    len(response['Parameters']).should.equal(10)
    response['NextToken'].should.equal('10')

    response = client.describe_parameters(NextToken=response['NextToken'])
    len(response['Parameters']).should.equal(10)
    response['NextToken'].should.equal('20')

    response = client.describe_parameters(NextToken=response['NextToken'])
    len(response['Parameters']).should.equal(10)
    response['NextToken'].should.equal('30')

    response = client.describe_parameters(NextToken=response['NextToken'])
    len(response['Parameters']).should.equal(10)
    response['NextToken'].should.equal('40')

    response = client.describe_parameters(NextToken=response['NextToken'])
    len(response['Parameters']).should.equal(10)
    response['NextToken'].should.equal('50')

    response = client.describe_parameters(NextToken=response['NextToken'])
    len(response['Parameters']).should.equal(0)
    ''.should.equal(response.get('NextToken', ''))


@mock_ssm
def test_describe_parameters_filter_names():
    client = boto3.client('ssm', region_name='us-east-1')

    for i in range(50):
        p = {
            'Name': "param-%d" % i,
            'Value': "value-%d" % i,
            'Type': "String"
        }
        if i % 5 == 0:
            p['Type'] = 'SecureString'
            p['KeyId'] = 'a key'
        client.put_parameter(**p)

    response = client.describe_parameters(Filters=[
        {
            'Key': 'Name',
            'Values': ['param-22']
        },
    ])
    len(response['Parameters']).should.equal(1)
    response['Parameters'][0]['Name'].should.equal('param-22')
    response['Parameters'][0]['Type'].should.equal('String')
    ''.should.equal(response.get('NextToken', ''))


@mock_ssm
def test_describe_parameters_filter_type():
    client = boto3.client('ssm', region_name='us-east-1')

    for i in range(50):
        p = {
            'Name': "param-%d" % i,
            'Value': "value-%d" % i,
            'Type': "String"
        }
        if i % 5 == 0:
            p['Type'] = 'SecureString'
            p['KeyId'] = 'a key'
        client.put_parameter(**p)

    response = client.describe_parameters(Filters=[
        {
            'Key': 'Type',
            'Values': ['SecureString']
        },
    ])
    len(response['Parameters']).should.equal(10)
    response['Parameters'][0]['Type'].should.equal('SecureString')
    '10'.should.equal(response.get('NextToken', ''))


@mock_ssm
def test_describe_parameters_filter_keyid():
    client = boto3.client('ssm', region_name='us-east-1')

    for i in range(50):
        p = {
            'Name': "param-%d" % i,
            'Value': "value-%d" % i,
            'Type': "String"
        }
        if i % 5 == 0:
            p['Type'] = 'SecureString'
            p['KeyId'] = "key:%d" % i
        client.put_parameter(**p)

    response = client.describe_parameters(Filters=[
        {
            'Key': 'KeyId',
            'Values': ['key:10']
        },
    ])
    len(response['Parameters']).should.equal(1)
    response['Parameters'][0]['Name'].should.equal('param-10')
    response['Parameters'][0]['Type'].should.equal('SecureString')
    ''.should.equal(response.get('NextToken', ''))


@mock_ssm
def test_get_parameter_invalid():
    client = client = boto3.client('ssm', region_name='us-east-1')
    response = client.get_parameters(
        Names=[
            'invalid'
        ],
        WithDecryption=False)

    len(response['Parameters']).should.equal(0)
    len(response['InvalidParameters']).should.equal(1)
    response['InvalidParameters'][0].should.equal('invalid')


@mock_ssm
def test_put_parameter_secure_default_kms():
    client = boto3.client('ssm', region_name='us-east-1')

    client.put_parameter(
        Name='test',
        Description='A test parameter',
        Value='value',
        Type='SecureString')

    response = client.get_parameters(
        Names=[
            'test'
        ],
        WithDecryption=False)

    len(response['Parameters']).should.equal(1)
    response['Parameters'][0]['Name'].should.equal('test')
    response['Parameters'][0]['Value'].should.equal('kms:default:value')
    response['Parameters'][0]['Type'].should.equal('SecureString')

    response = client.get_parameters(
        Names=[
            'test'
        ],
        WithDecryption=True)

    len(response['Parameters']).should.equal(1)
    response['Parameters'][0]['Name'].should.equal('test')
    response['Parameters'][0]['Value'].should.equal('value')
    response['Parameters'][0]['Type'].should.equal('SecureString')


@mock_ssm
def test_put_parameter_secure_custom_kms():
    client = boto3.client('ssm', region_name='us-east-1')

    client.put_parameter(
        Name='test',
        Description='A test parameter',
        Value='value',
        Type='SecureString',
        KeyId='foo')

    response = client.get_parameters(
        Names=[
            'test'
        ],
        WithDecryption=False)

    len(response['Parameters']).should.equal(1)
    response['Parameters'][0]['Name'].should.equal('test')
    response['Parameters'][0]['Value'].should.equal('kms:foo:value')
    response['Parameters'][0]['Type'].should.equal('SecureString')

    response = client.get_parameters(
        Names=[
            'test'
        ],
        WithDecryption=True)

    len(response['Parameters']).should.equal(1)
    response['Parameters'][0]['Name'].should.equal('test')
    response['Parameters'][0]['Value'].should.equal('value')
    response['Parameters'][0]['Type'].should.equal('SecureString')


@mock_ssm
def test_add_remove_list_tags_for_resource():
    client = boto3.client('ssm', region_name='us-east-1')

    client.add_tags_to_resource(
        ResourceId='test',
        ResourceType='Parameter',
        Tags=[{'Key': 'test-key', 'Value': 'test-value'}]
    )

    response = client.list_tags_for_resource(
        ResourceId='test',
        ResourceType='Parameter'
    )
    len(response['TagList']).should.equal(1)
    response['TagList'][0]['Key'].should.equal('test-key')
    response['TagList'][0]['Value'].should.equal('test-value')

    client.remove_tags_from_resource(
        ResourceId='test',
        ResourceType='Parameter',
        TagKeys=['test-key']
    )

    response = client.list_tags_for_resource(
        ResourceId='test',
        ResourceType='Parameter'
    )
    len(response['TagList']).should.equal(0)



from __future__ import unicode_literals
from .models import apigateway_backends
from ..core.models import base_decorator, deprecated_base_decorator

apigateway_backend = apigateway_backends['us-east-1']
mock_apigateway = base_decorator(apigateway_backends)
mock_apigateway_deprecated = deprecated_base_decorator(apigateway_backends)



from __future__ import unicode_literals

import bisect
import datetime
from collections import defaultdict
import json
from moto.core import BaseBackend, BaseModel
from moto.ec2 import ec2_backends
from .exceptions import BadSegmentException, AWSError


class TelemetryRecords(BaseModel):
    def __init__(self, instance_id, hostname, resource_arn, records):
        self.instance_id = instance_id
        self.hostname = hostname
        self.resource_arn = resource_arn
        self.records = records

    @classmethod
    def from_json(cls, json):
        instance_id = json.get('EC2InstanceId', None)
        hostname = json.get('Hostname')
        resource_arn = json.get('ResourceARN')
        telemetry_records = json['TelemetryRecords']

        return cls(instance_id, hostname, resource_arn, telemetry_records)


# https://docs.aws.amazon.com/xray/latest/devguide/xray-api-segmentdocuments.html
class TraceSegment(BaseModel):
    def __init__(self, name, segment_id, trace_id, start_time, raw, end_time=None, in_progress=False, service=None, user=None,
                 origin=None, parent_id=None, http=None, aws=None, metadata=None, annotations=None, subsegments=None, **kwargs):
        self.name = name
        self.id = segment_id
        self.trace_id = trace_id
        self._trace_version = None
        self._original_request_start_time = None
        self._trace_identifier = None
        self.start_time = start_time
        self._start_date = None
        self.end_time = end_time
        self._end_date = None
        self.in_progress = in_progress
        self.service = service
        self.user = user
        self.origin = origin
        self.parent_id = parent_id
        self.http = http
        self.aws = aws
        self.metadata = metadata
        self.annotations = annotations
        self.subsegments = subsegments
        self.misc = kwargs

        # Raw json string
        self.raw = raw

    def __lt__(self, other):
        return self.start_date < other.start_date

    @property
    def trace_version(self):
        if self._trace_version is None:
            self._trace_version = int(self.trace_id.split('-', 1)[0])
        return self._trace_version

    @property
    def request_start_date(self):
        if self._original_request_start_time is None:
            start_time = int(self.trace_id.split('-')[1], 16)
            self._original_request_start_time = datetime.datetime.fromtimestamp(start_time)
        return self._original_request_start_time

    @property
    def start_date(self):
        if self._start_date is None:
            self._start_date = datetime.datetime.fromtimestamp(self.start_time)
        return self._start_date

    @property
    def end_date(self):
        if self._end_date is None:
            self._end_date = datetime.datetime.fromtimestamp(self.end_time)
        return self._end_date

    @classmethod
    def from_dict(cls, data, raw):
        # Check manditory args
        if 'id' not in data:
            raise BadSegmentException(code='MissingParam', message='Missing segment ID')
        seg_id = data['id']
        data['segment_id'] = seg_id  # Just adding this key for future convenience

        for arg in ('name', 'trace_id', 'start_time'):
            if arg not in data:
                raise BadSegmentException(seg_id=seg_id, code='MissingParam', message='Missing segment ID')

        if 'end_time' not in data and 'in_progress' not in data:
            raise BadSegmentException(seg_id=seg_id, code='MissingParam', message='Missing end_time or in_progress')
        if 'end_time' not in data and data['in_progress'] == 'false':
            raise BadSegmentException(seg_id=seg_id, code='MissingParam', message='Missing end_time')

        return cls(raw=raw, **data)


class SegmentCollection(object):
    def __init__(self):
        self._traces = defaultdict(self._new_trace_item)

    @staticmethod
    def _new_trace_item():
        return {
            'start_date': datetime.datetime(1970, 1, 1),
            'end_date': datetime.datetime(1970, 1, 1),
            'finished': False,
            'trace_id': None,
            'segments': []
        }

    def put_segment(self, segment):
        # insert into a sorted list
        bisect.insort_left(self._traces[segment.trace_id]['segments'], segment)

        # Get the last segment (takes into account incorrect ordering)
        # and if its the last one, mark trace as complete
        if self._traces[segment.trace_id]['segments'][-1].end_time is not None:
            self._traces[segment.trace_id]['finished'] = True

            start_time = self._traces[segment.trace_id]['segments'][0].start_date
            end_time = self._traces[segment.trace_id]['segments'][-1].end_date
            self._traces[segment.trace_id]['start_date'] = start_time
            self._traces[segment.trace_id]['end_date'] = end_time
            self._traces[segment.trace_id]['trace_id'] = segment.trace_id
            # Todo consolidate trace segments into a trace.
            # not enough working knowledge of xray to do this

    def summary(self, start_time, end_time, filter_expression=None, sampling=False):
        # This beast https://docs.aws.amazon.com/xray/latest/api/API_GetTraceSummaries.html#API_GetTraceSummaries_ResponseSyntax
        if filter_expression is not None:
            raise AWSError('Not implemented yet - moto', code='InternalFailure', status=500)

        summaries = []

        for tid, trace in self._traces.items():
            if trace['finished'] and start_time < trace['start_date'] and trace['end_date'] < end_time:
                duration = int((trace['end_date'] - trace['start_date']).total_seconds())
                # this stuff is mostly guesses, refer to TODO above
                has_error = any(['error' in seg.misc for seg in trace['segments']])
                has_fault = any(['fault' in seg.misc for seg in trace['segments']])
                has_throttle = any(['throttle' in seg.misc for seg in trace['segments']])

                # Apparently all of these options are optional
                summary_part = {
                    'Annotations': {},  # Not implemented yet
                    'Duration': duration,
                    'HasError': has_error,
                    'HasFault': has_fault,
                    'HasThrottle': has_throttle,
                    'Http': {},  # Not implemented yet
                    'Id': tid,
                    'IsParital': False,  # needs lots more work to work on partials
                    'ResponseTime': 1,  # definitely 1ms resposnetime
                    'ServiceIds': [],  # Not implemented yet
                    'Users': {}  # Not implemented yet
                }
                summaries.append(summary_part)

        result = {
            "ApproximateTime": int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds()),
            "TracesProcessedCount": len(summaries),
            "TraceSummaries": summaries
        }

        return result

    def get_trace_ids(self, trace_ids):
        traces = []
        unprocessed = []

        # Its a default dict
        existing_trace_ids = list(self._traces.keys())
        for trace_id in trace_ids:
            if trace_id in existing_trace_ids:
                traces.append(self._traces[trace_id])
            else:
                unprocessed.append(trace_id)

        return traces, unprocessed


class XRayBackend(BaseBackend):

    def __init__(self):
        self._telemetry_records = []
        self._segment_collection = SegmentCollection()

    def add_telemetry_records(self, json):
        self._telemetry_records.append(
            TelemetryRecords.from_json(json)
        )

    def process_segment(self, doc):
        try:
            data = json.loads(doc)
        except ValueError:
            raise BadSegmentException(code='JSONFormatError', message='Bad JSON data')

        try:
            # Get Segment Object
            segment = TraceSegment.from_dict(data, raw=doc)
        except ValueError:
            raise BadSegmentException(code='JSONFormatError', message='Bad JSON data')

        try:
            # Store Segment Object
            self._segment_collection.put_segment(segment)
        except Exception as err:
            raise BadSegmentException(seg_id=segment.id, code='InternalFailure', message=str(err))

    def get_trace_summary(self, start_time, end_time, filter_expression, summaries):
        return self._segment_collection.summary(start_time, end_time, filter_expression, summaries)

    def get_trace_ids(self, trace_ids, next_token):
        traces, unprocessed_ids = self._segment_collection.get_trace_ids(trace_ids)

        result = {
            'Traces': [],
            'UnprocessedTraceIds': unprocessed_ids

        }

        for trace in traces:
            segments = []
            for segment in trace['segments']:
                segments.append({
                    'Id': segment.id,
                    'Document': segment.raw
                })

            result['Traces'].append({
                'Duration': int((trace['end_date'] - trace['start_date']).total_seconds()),
                'Id': trace['trace_id'],
                'Segments': segments
            })

        return result


xray_backends = {}
for region, ec2_backend in ec2_backends.items():
    xray_backends[region] = XRayBackend()



import sure  # noqa

from moto.swf.utils import decapitalize


def test_decapitalize():
    cases = {
        "fooBar": "fooBar",
        "FooBar": "fooBar",
        "FOO BAR": "fOO BAR",
    }
    for before, after in cases.items():
        decapitalize(before).should.equal(after)
